import os
import shutil
import hashlib
import fnmatch

from urllib.request import urlopen, HTTPError
import logging
from contextlib import contextmanager
from pysnptools.util.filecache import FileCache
import tempfile
import json
from pysnptools.util import log_in_place

# Based on an idea see in Danilo Horta's Bgen-reader-py
class Hashdown(FileCache):
    """
    A :class:`.FileCache` for downloading files with known MD5 hashes from a URL. Unloading is not supported.

    See :class:`.FileCache` for general examples of using FileCache.

    **Constructor:**
        :Parameters: * **url** (*string*) -- The URL from which to download any needed files.
                     * **file_to_hash** (*dictionary*) -- A dictionary from file names to MD5 hashes.
                     * **directory** (optional, *string*) -- Local location for files. If not given will be under the system temp directory
                            (typically controlled with the TEMP environment variable).
                     * **allow_unknown_files** (optional, *bool*) -- By default, all requested files must be in the dictionary. If True,
                            other files can be requested. If found under the URL, they will be downloaded and an entry will be added
                            to the dictionary.
                     * **trust_local_files** (optional, bool) -- By default, when **allow_unknown_files** is True, unknown files
                            will be download. If **trust_local_files** is also True, then any local files in **directory** will
                            be assumed to have the correct hash.
                     * **_relative_directory** (*string*) -- Used internally

        :Example:

        >>> from pysnptools.util.filecache import Hashdown
        >>> file_to_hash= {'pysnptools/examples/toydata.5chrom.bed': '766f55aa716bc7bc97cad4de41a50ec3',
        ...                'pysnptools/examples/toydata.5chrom.bim': '6a07f96e521f9a86df7bfd7814eebcd6',
        ...                'pysnptools/examples/toydata.5chrom.fam': 'f4eb01f67e0738d4865fad2014af8537'}
        >>> hashdown = Hashdown('https://github.com/fastlmm/PySnpTools/raw/cf248cbf762516540470d693532590a77c76fba2',
        ...                      file_to_hash=file_to_hash)
        >>> hashdown.file_exists('pysnptools/examples/toydata.5chrom.bed')
        True
        >>> hashdown.load('pysnptools/examples/toydata.5chrom.fam').split('\\n')[0]
        'per0 per0 0 0 2 0.408848'

    """

    def __init__(
        self,
        url,
        file_to_hash={},
        directory=None,
        allow_unknown_files=False,
        trust_local_files=False,
        _relative_directory=None,
    ):
        super(Hashdown, self).__init__()
        self.url = url
        self.file_to_hash = file_to_hash
        self.allow_unknown_files = allow_unknown_files
        self.trust_local_files = trust_local_files
        base_url = (
            url if _relative_directory is None else url[: -len(_relative_directory) - 1]
        )
        url_hash = hashlib.md5(base_url.encode("utf-8")).hexdigest()
        self.directory = tempfile.gettempdir() + "/hashdown/{0}".format(url_hash)
        self._relative_directory = _relative_directory
        if os.path.exists(self.directory):
            assert not os.path.isfile(
                self.directory
            ), "A directory cannot exist where a file already exists."

    def __repr__(self):
        return "{0}('{1}')".format(self.__class__.__name__, self.url)

    @property
    def name(self):
        """
        A path-like name for this `Hashdown`.

        :rtype: string

        >>> from pysnptools.util.filecache import Hashdown
        >>> file_to_hash= {'pysnptools/examples/toydata.5chrom.bed': '766f55aa716bc7bc97cad4de41a50ec3',
        ...                'pysnptools/examples/toydata.5chrom.bim': '6a07f96e521f9a86df7bfd7814eebcd6',
        ...                'pysnptools/examples/toydata.5chrom.fam': 'f4eb01f67e0738d4865fad2014af8537'}
        >>> hashdown = Hashdown('https://github.com/fastlmm/PySnpTools/raw/cf248cbf762516540470d693532590a77c76fba2',
        ...                      file_to_hash=file_to_hash)
        >>> hashdown.name
        'hashdown/9ac30da2bf589db947e91744dff0ec24'
        >>> hashdown.join('pysnptools').name
        'hashdown/9ac30da2bf589db947e91744dff0ec24/pysnptools'

        """
        if self._relative_directory is None:
            url_hash = hashlib.md5(self.url.encode("utf-8")).hexdigest()
            result = "hashdown/{0}".format(url_hash)
        else:
            base_url = self.url[: -len(self._relative_directory) - 1]
            url_hash = hashlib.md5(base_url.encode("utf-8")).hexdigest()
            result = "hashdown/{0}/{1}".format(url_hash, self._relative_directory)

        return result

    @staticmethod
    def _get_large_file(url, file, trust_local_files, length=16 * 1024):
        """https://stackoverflow.com/questions/1517616/stream-large-binary-files-with-urllib2-to-file
        """
        import pysnptools.util as pstutil # put here to avoid recursive nesting

        logging.info("Downloading'{0}'".format(url))
        if trust_local_files and os.path.exists(file):
            logging.info("Trusting local file'{0}'".format(file))
            return True

        try:
            req = urlopen(url)
        except HTTPError as e:
            if e.code == 404:
                return False
            raise

        pstutil.create_directory_if_necessary(file)
        with open(file, "wb") as fp:
            shutil.copyfileobj(req, fp, length)
        return True

    @staticmethod
    def _get_hash(filename):
        """https://stackoverflow.com/questions/16874598/how-do-i-calculate-the-md5-checksum-of-a-file-in-python
        """
        logging.info("Find hash of '{0}'".format(filename))
        with open(filename, "rb") as f:
            file_hash = hashlib.md5()
            while True:
                chunk = f.read(8192)
                if chunk == b"":
                    break
                file_hash.update(chunk)
        return file_hash.hexdigest()

    def _simple_file_exists(self, simple_file_name):
        rel_part = (
            "" if self._relative_directory is None else self._relative_directory + "/"
        )
        full_file = self.directory + "/" + rel_part + simple_file_name
        relative_file = (
            simple_file_name
            if self._relative_directory is None
            else self._relative_directory + "/" + simple_file_name
        )
        full_url = self.url + "/" + simple_file_name
        hash = self.file_to_hash.get(relative_file)

        if hash is not None:
            return True

        if not self.allow_unknown_files:
            return False

        if self._get_large_file(full_url, full_file, self.trust_local_files):
            hash = Hashdown._get_hash(full_file)
            self.file_to_hash[relative_file] = hash
            return True
        else:
            return False

    def _example_file(self, pattern, endswith=None):
        return_file = None
        for filename in fnmatch.filter(self.file_to_hash, pattern):
            with self.open_read(filename) as local_file:
                if return_file is None and (
                    endswith is None or fnmatch.fnmatch(filename, endswith)
                ):
                    return_file = local_file
        assert return_file is not None, "Pattern not found '{0}'{1}".format(
            pattern, "" if (endswith is None) else "'{0}'".format(endswith)
        )
        return return_file

    @contextmanager
    def _simple_open_read(self, simple_file_name, updater=None):
        logging.info("open_read('{0}')".format(simple_file_name))

        relative_file = (
            simple_file_name
            if self._relative_directory is None
            else self._relative_directory + "/" + simple_file_name
        )
        full_file = self.directory + "/" + relative_file
        full_url = self.url + "/" + simple_file_name
        hash = self.file_to_hash.get(relative_file)
        assert self._simple_file_exists(
            simple_file_name
        ), "File doesn't exist ('{0}')".format(relative_file)
        if os.path.exists(full_file):
            local_hash = Hashdown._get_hash(full_file)
        else:
            local_hash = None

        if local_hash is None or local_hash != hash:
            assert self._get_large_file(
                full_url, full_file, trust_local_files=False
            ), "URL return 'no item' ('{0}')".format(full_url)
            local_hash = Hashdown._get_hash(full_file)
            if hash is None:
                assert self.allow_unknown_files, "real assert"
                self.file_to_hash[relative_file] = local_hash
            else:
                assert (
                    hash == local_hash
                ), 'URL file has unexpected hash ("{0}")'.format(full_url)

        yield full_file

        logging.info("close('{0}')".format(simple_file_name))

    @contextmanager
    def _simple_open_write(self, simple_file_name, size=0, updater=None):
        raise ValueError("Hashdown is read only. writing is not allowed.")

    def _simple_rmtree(self, updater=None):
        raise ValueError('Hashdown is read only. "rmtree" is not allowed.')

    def _simple_remove(self, simple_file_name, updater=None):
        raise ValueError('Hashdown is read only. "remove" is not allowed.')

    def _simple_getmtime(self, simple_file_name):
        assert self._simple_file_exists(
            simple_file_name
        ), "file doesn't exist ('{0}')".format(simple_file_name)
        return 0

    def _simple_join(self, path):
        directory = self.directory + "/" + path
        _relative_directory = (
            path
            if self._relative_directory is None
            else self._relative_directory + "/" + path
        )
        if not self.allow_unknown_files:
            assert not self.file_exists(
                _relative_directory
            ), "Can't treat an existing file as a directory"
        return Hashdown(
            url=self.url + "/" + path,
            directory=directory,
            file_to_hash=self.file_to_hash,
            allow_unknown_files=self.allow_unknown_files,
            _relative_directory=_relative_directory,
        )

    def _simple_walk(self):
        for rel_file in self.file_to_hash:
            if self._relative_directory is None or rel_file.startswith(
                self._relative_directory + "/"
            ):
                file = (
                    rel_file
                    if self._relative_directory is None
                    else rel_file[len(self._relative_directory) + 1:]
                )
                if self.file_exists(file):
                    yield file

    def save_hashdown(self, filename):
        """
        Save a Hashdown object to a json file.

        :param filename: name of file to save to.
        :type path: string

        >>> from pysnptools.util.filecache import Hashdown
        >>> file_to_hash= {'pysnptools/examples/toydata.5chrom.bed': '766f55aa716bc7bc97cad4de41a50ec3',
        ...                'pysnptools/examples/toydata.5chrom.bim': '6a07f96e521f9a86df7bfd7814eebcd6',
        ...                'pysnptools/examples/toydata.5chrom.fam': 'f4eb01f67e0738d4865fad2014af8537'}
        >>> hashdown1 = Hashdown('https://github.com/fastlmm/PySnpTools/raw/cf248cbf762516540470d693532590a77c76fba2',
        ...                      file_to_hash=file_to_hash)
        >>> hashdown1.save_hashdown('tempdir/demo.hashdown.json')
        >>> hashdown2 = Hashdown.load_hashdown('tempdir/demo.hashdown.json')
        >>> hashdown2.file_exists('pysnptools/examples/toydata.5chrom.bed')
        True
        >>> hashdown2.load('pysnptools/examples/toydata.5chrom.fam').split('\\n')[0]
        'per0 per0 0 0 2 0.408848'

        """
        import pysnptools.util as pstutil # put here to avoid recursive nesting

        pstutil.create_directory_if_necessary(filename)
        dict0 = dict(self.__dict__)
        del dict0["directory"]
        del dict0["_relative_directory"]
        del dict0["allow_unknown_files"]
        del dict0["trust_local_files"]
        with open(filename, "w") as json_file:
            json.dump(dict0, json_file)

    @staticmethod
    def load_hashdown(
        filename, directory=None, allow_unknown_files=False, trust_local_files=False
    ):
        """
        Load a Hashdown object from a json file.

        :param filename: name of file to load from.
        :type path: string

        :param directory:  Local location for files. If not given will be under the system temp directory
                           (typically controlled with the TEMP environment variable).
        :type path: string

        :param allow_unknown_files: By default, all requested files must be in the dictionary. If True,
                                    other files can be requested. If found under the URL, they will be downloaded and an entry will be added
                                    to the dictionary.
        :type path: bool

        :param trust_local_files: By default, when **allow_unknown_files** is True, unknown files
                                  will be download. If **trust_local_files** is also True, then any local files in **directory** will
                                  be assumed to have the correct hash.
        :type path: bool

        :rtype: :class:`.Hashdown`

        >>> from pysnptools.util.filecache import Hashdown
        >>> file_to_hash= {'pysnptools/examples/toydata.5chrom.bed': '766f55aa716bc7bc97cad4de41a50ec3',
        ...                'pysnptools/examples/toydata.5chrom.bim': '6a07f96e521f9a86df7bfd7814eebcd6',
        ...                'pysnptools/examples/toydata.5chrom.fam': 'f4eb01f67e0738d4865fad2014af8537'}
        >>> hashdown1 = Hashdown('https://github.com/fastlmm/PySnpTools/raw/cf248cbf762516540470d693532590a77c76fba2',
        ...                      file_to_hash=file_to_hash)
        >>> hashdown1.save_hashdown('tempdir/demo.hashdown.json')
        >>> hashdown2 = Hashdown.load_hashdown('tempdir/demo.hashdown.json')
        >>> hashdown2.file_exists('pysnptools/examples/toydata.5chrom.bed')
        True
        >>> hashdown2.load('pysnptools/examples/toydata.5chrom.fam').split('\\n')[0]
        'per0 per0 0 0 2 0.408848'

        """
        with open(filename) as json_file:
            dict0 = json.load(json_file)
        hashdown = Hashdown(
            url=dict0["url"],
            file_to_hash=dict0["file_to_hash"],
            directory=directory,
            allow_unknown_files=allow_unknown_files,
            trust_local_files=trust_local_files,
        )
        return hashdown

    @staticmethod
    def scan_local(local_directory, url=None, logging_level=logging.WARNING):
        '''
        Bootstrap a Hashdown by recursively walking a local directory and finding the local MD5 hashes.
        (A local hash might be wrong if the files are out of date or have OS-dependent line endings.)
        Typically, you'll then want to save the result to a JSON file and then edit that JSON file
        manually to remove uninteresting files.

        :param local_directory: Local directory to recursively walk
        :type path: string

        :param url:  URL to give to the Hashdown. (It will not be checked.)
        :type path: string

        :param logging_level: Logging level for printing progress of the walk. Default
               is logging.WARNING)

        :rtype: :class:`.Hashdown`

        '''
        from pysnptools.util.filecache import LocalCache

        file_to_hash = {}
        localcache = LocalCache(local_directory)
        with log_in_place("scanning", logging_level) as updater:
            for file in localcache.walk():
                updater(file)
                with localcache.open_read(file) as full_file:
                    hash = Hashdown._get_hash(full_file)
                    file_to_hash[file] = hash
        return Hashdown(url, file_to_hash=file_to_hash)


if __name__ == "__main__":
    import doctest
    if False:
        hashdown_local = Hashdown.scan_local(
            r"D:\OneDrive\programs\fastlmm",
            url="https://github.com/fastlmm/FaST-LMM/raw/ff183d3aa09c78cf5fdf1961e9241e8a9b9dd172",
        )
        hashdown_local.save_hashdown("deldir/fastlmm.hashdown.json")

    logging.basicConfig(level=logging.INFO)

    doctest.testmod(optionflags=doctest.ELLIPSIS)



import numpy as np
from itertools import *
import pandas as pd
import logging
from pysnptools.standardizer import Unit, Standardizer, UnitTrained, BetaTrained
from pysnptools.standardizer import Identity as SS_Identity
from pysnptools.kernelreader import KernelReader
from pysnptools.kernelreader import KernelData
from pysnptools.kernelstandardizer import DiagKtoN
import pysnptools.util as pstutil


class SnpKernel(KernelReader):
    """
    A :class:`.KernelReader` that creates a kernel from a :class:`.SnpReader`. No SNP data will be read until
    the :meth:`SnpKernel.read` method is called. Use block_size to avoid ever reading all the SNP data into memory.
    at once.

    See :class:`.KernelReader` for general examples of using KernelReaders.

    **Constructor:**
        :Parameters: * **snpreader** (:class:`SnpReader`) -- The SNP data
                     * **standardizer** (:class:`Standardizer`) -- How the SNP data should be standardized
                     * **block_size** (optional, int) -- The number of SNPs to read at a time.

        If **block_size** is not given, then all SNP data will be read at once.

        :Example:

        >>> from pysnptools.snpreader import Bed
        >>> from pysnptools.standardizer import Unit
        >>> from pysnptools.util import example_file # Download and return local file name
        >>> bed_file = example_file('pysnptools/examples/toydata.5chrom.*','*.bed')
        >>> snp_on_disk = Bed(bed_file,count_A1=False)     # A Bed file is specified, but nothing is read from disk
        >>> kernel_on_disk = SnpKernel(snp_on_disk, Unit(),block_size=500)  # A kernel is specified, but nothing is read from disk
        >>> print(kernel_on_disk) #Print the specification
        SnpKernel(Bed(...pysnptools/examples/toydata.5chrom.bed',count_A1=False),standardizer=Unit(),block_size=500)
        >>> print(kernel_on_disk.iid_count)                                  # iid information is read from disk, but not SNP data
        500
        >>> kerneldata = kernel_on_disk.read().standardize()                # SNPs are read and Unit standardized, 500 at a time, to create a kernel, which is then standardized
        >>> print('{0:.6f}'.format(kerneldata.val[0,0]))
        0.992307
    """

    def __init__(self, snpreader, standardizer=None, block_size=None):
        super(SnpKernel, self).__init__()

        assert standardizer is not None, "'standardizer' must be provided"

        self.snpreader = snpreader
        self.standardizer = standardizer
        self.block_size = block_size

    @property
    def row(self):
        return self.snpreader.iid

    @property
    def col(self):
        if hasattr(self, "_col"):
            return self._col
        else:
            return self.snpreader.iid

    def __repr__(self):
        return self._internal_repr(self.standardizer)

    def _internal_repr(self, standardizer):  # !!! merge this with __repr__
        s = "SnpKernel({0},standardizer={1}".format(self.snpreader, standardizer)
        if self.block_size is not None:
            s += ",block_size={0}".format(self.block_size)
        s += ")"
        return s

    def copyinputs(self, copier):
        # Doesn't need run_once
        copier.input(self.snpreader)
        copier.input(self.standardizer)

    def _read(
        self,
        row_index_or_none,
        col_index_or_none,
        order,
        dtype,
        force_python_only,
        view_ok,
        num_threads,
    ):
        dtype = np.dtype(dtype)
        # Special case: If square and constant, can push the subsetting into the SnpReader
        if (
            self.standardizer.is_constant
            and row_index_or_none is not None
            and col_index_or_none is not None
            and np.array_equal(row_index_or_none, col_index_or_none)
        ):
            return self.snpreader[row_index_or_none, :]._read_kernel(
                self.standardizer,
                self.block_size,
                order,
                dtype,
                force_python_only,
                view_ok,
                num_threads,
            )  # !!!LATER can this ever be reached?
        else:
            # LATER: If it was often that case that we wanted to standardize on all the data, but then only return a slice of the result,
            #       that could be done with less memory by working in blocks but not tabulating for all the iids.
            whole = self.snpreader._read_kernel(
                self.standardizer,
                self.block_size,
                order,
                dtype,
                force_python_only,
                view_ok,
                num_threads,
            )
            val, shares_memory = self._apply_sparray_or_slice_to_val(
                whole,
                row_index_or_none,
                col_index_or_none,
                order,
                dtype,
                force_python_only,
                num_threads,
            )
            return val

    def __getitem__(self, iid_indexer_and_snp_indexer):
        if isinstance(iid_indexer_and_snp_indexer, tuple):
            row_index_or_none, col_index_or_none = iid_indexer_and_snp_indexer
        else:
            row_index_or_none = iid_indexer_and_snp_indexer
            col_index_or_none = row_index_or_none

        # Special case: If square and constant, can push the subsetting into the SnpReader
        if (
            self.standardizer.is_constant
            and row_index_or_none is not None
            and col_index_or_none is not None
            and np.array_equal(row_index_or_none, col_index_or_none)
        ):
            return SnpKernel(
                self.snpreader[row_index_or_none, :],
                self.standardizer,
                block_size=self.block_size,
            )
        else:
            return KernelReader.__getitem__(self, iid_indexer_and_snp_indexer)

    def _read_with_standardizing(
        self,
        to_kerneldata,
        kernel_standardizer=DiagKtoN(),
        return_trained=False,
        num_threads=None,
    ):
        """
        Reads a SnpKernel with two cases
              If returning KernelData,
                 just calls snpreader._read_kernel, package, kernel_standardize
              If returning simple SnpKernel that needs no more standardization
                  read the reference and learn both standardization (but can't this cause multiple reads?)
        Note that snp_standardizer should be None or the standardizer instead the SnpKernel should have the placeholder value Standardizer()

        Will choose which array module to use based on the ARRAY_MODULE environment variable (e.g. 'numpy' (default) or 'cupy')

        """
        logging.info("Starting '_read_with_standardizing'")
        xp = pstutil.array_module()

        if to_kerneldata:
            val, snp_trained = self.snpreader._read_kernel(
                self.standardizer,
                block_size=self.block_size,
                return_trained=True,
                num_threads=num_threads,
            )
            kernel = KernelData(iid=self.snpreader.iid, val=val, name=str(self), xp=xp)
            kernel, kernel_trained = kernel.standardize(
                kernel_standardizer, return_trained=True, num_threads=num_threads
            )
        else:
            snpdata, snp_trained = self.snpreader.read().standardize(
                self.standardizer, return_trained=True, num_threads=num_threads
            )
            snpdata, kernel_trained = snpdata.standardize(
                kernel_standardizer, return_trained=True, num_threads=num_threads
            )
            kernel = SnpKernel(snpdata, SS_Identity())

        logging.info("Ending '_read_with_standardizing'")
        if return_trained:
            return kernel, snp_trained, kernel_trained
        else:
            return kernel

    @property
    def sid(self):
        """The :attr:`SnpReader.sid` property of the SNP data."""
        return self.snpreader.sid

    @property
    def sid_count(self):
        """The :attr:`SnpReader.sid_count` property of the SNP data."""
        return self.snpreader.sid_count

    @property
    def pos(self):
        """The :attr:`SnpReader.pos` property of the SNP data."""
        return self.snpreader.pos

    # LATER this needs a test
    def read_snps(
        self,
        order="F",
        dtype=np.float64,
        force_python_only=False,
        view_ok=False,
        num_threads=None,
    ):
        """Reads the SNP values, applies the standardizer, and returns a :class:`.SnpData`.

        :param order: {'F' (default), 'C', 'A'}, optional -- Specify the order of the ndarray. If order is 'F' (default),
            then the array will be in F-contiguous order (iid-index varies the fastest).
            If order is 'C', then the returned array will be in C-contiguous order (sid-index varies the fastest).
            If order is 'A', then the :attr:`SnpData.val`
            ndarray may be in any order (either C-, Fortran-contiguous).
        :type order: string or None

        :param dtype: {numpy.float64 (default), numpy.float32}, optional -- The data-type for the :attr:`SnpData.val` ndarray.
        :type dtype: data-type

        :param force_python_only: optional -- If False (default), may use outside library code. If True, requests that the read
            be done without outside library code.
        :type force_python_only: bool

        :param view_ok: optional -- If False (default), allocates new memory for the :attr:`SnpData.val`'s ndarray. If True,
            if practical and reading from a :class:`SnpData`, will return a new
            :class:`SnpData` with a ndarray shares memory with the original :class:`SnpData`.
            Typically, you'll also wish to use "order='A'" to increase the chance that sharing will be possible.
            Use these parameters with care because any change to either ndarray (for example, via :meth:`.SnpData.standardize`) will effect
            the others. Also keep in mind that :meth:`read` relies on ndarray's mechanisms to decide whether to actually
            share memory and so it may ignore your suggestion and allocate a new ndarray anyway.
        :type view_ok: bool

        :param num_threads: optional -- The number of threads with which to read data. Defaults to all available
            processors. Can also be set with these environment variables (listed in priority order):
            'PST_NUM_THREADS', 'NUM_THREADS', 'MKL_NUM_THREADS'.
        :type num_threads: None or int

        :rtype: :class:`.SnpData`

        """
        dtype = np.dtype(dtype)
        return self.snpreader.read(
            order=order,
            dtype=dtype,
            force_python_only=force_python_only,
            view_ok=view_ok,
            num_threads=num_threads,
        ).standardize(self.standardizer, num_threads=num_threads)


if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO)

    import doctest

    doctest.testmod(optionflags=doctest.ELLIPSIS)
    # There is also a unit test case in 'pysnptools\test.py' that calls this doc test



import logging
import numpy as np
from pysnptools.snpreader import SnpReader
from pysnptools.pstreader import PstHdf5
import warnings


class SnpHdf5(PstHdf5, SnpReader):
    """
    A :class:`.SnpReader` for reading \*.snp.hdf5 files from disk.

    See :class:`.SnpReader` for general examples of using SnpReaders.

    The general HDF5 format is described `here <http://www.hdfgroup.org/HDF5/>`__. The SnpHdf5 format stores
    val, iid, sid, and pos information in Hdf5 format.

    **Constructor:**
        :Parameters: * **filename** (*string*) -- The SnpHdf5 file to read.

        :Example:

        >>> from pysnptools.snpreader import SnpHdf5
        >>> from pysnptools.util import example_file # Download and return local file name
        >>> hd5f_file = example_file("pysnptools/examples/toydata.snpmajor.snp.hdf5")
        >>> data_on_disk = SnpHdf5(hd5f_file)
        >>> print((data_on_disk.iid_count, data_on_disk.sid_count))
        (500, 10000)

    **Methods beyond** :class:`.SnpReader`

    """

    def __init__(self, *args, **kwargs):
        super(SnpHdf5, self).__init__(*args, **kwargs)

    @property
    def row(self):
        self._run_once()
        if self._row.dtype.type is not np.str_:
            self._row = np.array(self._row, dtype="str")
        return self._row

    @property
    def col(self):
        self._run_once()
        if self._col.dtype.type is not np.str_:
            self._col = np.array(self._col, dtype="str")
        return self._col

    # LATER make flush visible here and in other *.hdf5 locations

    @staticmethod
    def write(filename, snpdata, hdf5_dtype=None, sid_major=True):
        """Writes a :class:`SnpData` to SnpHdf5 format and return a the :class:`.SnpHdf5`.

        :param filename: the name of the file to create
        :type filename: string
        :param snpdata: The in-memory data that should be written to disk.
        :type snpdata: :class:`SnpData`
        :param hdf5_dtype: None (use the .val's dtype) or a Hdf5 dtype, e.g. 'f8','f4',etc.
        :type hdf5_dtype: string
        :param sid_major: Tells if vals should be stored on disk in sid_major (default) or iid_major format.
        :type sid_major: bool
        :rtype: :class:`.SnpHdf5`

        >>> from pysnptools.snpreader import SnpHdf5, Bed
        >>> import pysnptools.util as pstutil
        >>> from pysnptools.util import example_file # Download and return local file name
        >>> bedfile = example_file("pysnptools/examples/toydata.5chrom.*","*.bed")
        >>> snpdata = Bed(bedfile,count_A1=False)[:,:10].read()     # Read first 10 snps from Bed format
        >>> pstutil.create_directory_if_necessary("tempdir/toydata10.snp.hdf5")
        >>> SnpHdf5.write("tempdir/toydata10.snp.hdf5",snpdata)        # Write data in SnpHdf5 format
        SnpHdf5('tempdir/toydata10.snp.hdf5')
        """
        PstHdf5.write(filename, snpdata, hdf5_dtype=hdf5_dtype, col_major=sid_major)
        return SnpHdf5(filename)


class Hdf5(SnpHdf5):
    # !! warnings.warn("class 'Hdf5' is deprecated. Use the standard class 'SnpHdf5' instead", DeprecationWarning)
    def __init__(self, *args, **kwargs):
        super(Hdf5, self).__init__(*args, **kwargs)


if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO)

    import doctest

    doctest.testmod(optionflags=doctest.ELLIPSIS)



from pysnptools.util.mapreduce1.runner import (
    Runner,
    _JustCheckExists,
    _run_all_in_memory,
    _shape_to_desired_workcount,
    _work_sequence_for_one_index,
)
import os
import logging
import cloudpickle as pickle
import subprocess, sys, os.path
import threading
import pysnptools.util as util
from six.moves.queue import PriorityQueue


class LocalMultiThread(Runner):
    """
    A :class:`.Runner` that runs a :func:`.map_reduce` as multiple threads on a single machine.

    Note that Python has problems running some programs efficiently on multiple threads. (Search 'python global interpreter lock' for details.)
    If this :class:`.Runner` doesn't work, consider :class:`LocalMultiProc` instead.

    **Constructor:**
        :Parameters: * **taskcount** (*number*) -- The number of threads to run on.
        :Parameters: * **mkl_num_threads** (*number*) -- (default None) Limit on the number threads used by the NumPy MKL library.
        :Parameters: * **just_one_process** (*bool*) -- (default False) Divide the work for multiple threads, but sequentially on one thread. Can be useful for debugging.

        :Example:

        >>> from pysnptools.util.mapreduce1 import map_reduce
        >>> from pysnptools.util.mapreduce1.runner import LocalMultiThread
        >>> def holder1(n,runner):
        ...     def mapper1(x):
        ...         return x*x
        ...     def reducer1(sequence):
        ...        return sum(sequence)
        ...     return map_reduce(range(n),mapper=mapper1,reducer=reducer1,runner=runner)
        >>> holder1(100,LocalMultiThread(4))
        328350

    """

    def __init__(
        self,
        taskcount,
        mkl_num_threads=None,
        just_one_process=False,
    ):
        if not 0 < taskcount:
            raise Exception("Expect taskcount to be positive")

        self.taskcount = taskcount
        self.just_one_process = just_one_process
        if mkl_num_threads is not None:
            os.environ["MKL_NUM_THREADS"] = str(mkl_num_threads)

    def _result_sequence(self, thread_list, priority_queue, shaped_distributable):
        for thread in thread_list:
            if not self.just_one_process:
                thread.join()
            result_sequence = priority_queue.get()[1]
            for result in result_sequence:
                yield result

    def run(self, distributable):
        _JustCheckExists().input(distributable)

        priority_queue = PriorityQueue()
        thread_list = []
        shaped_distributable = _shape_to_desired_workcount(
            distributable, self.taskcount
        )
        for taskindex in range(self.taskcount):

            def _target(taskindex=taskindex):
                result_list = []
                for work in _work_sequence_for_one_index(
                    shaped_distributable, self.taskcount, taskindex
                ):
                    result_list.append(_run_all_in_memory(work))
                priority_queue.put((taskindex, result_list))

            if not self.just_one_process:
                thread = threading.Thread(target=_target, name=str(taskindex))
                thread_list.append(thread)
                thread.start()
            else:
                thread_list.append(None)
                _target()

        result_sequence = self._result_sequence(
            thread_list, priority_queue, shaped_distributable
        )
        result = shaped_distributable.reduce(result_sequence)

        _JustCheckExists().output(distributable)
        return result


if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO)

    import doctest

    doctest.testmod(optionflags=doctest.ELLIPSIS)



from pysnptools.snpreader import SnpReader
from pysnptools.pstreader._subset import _PstSubset

class _SnpSubset(_PstSubset,SnpReader):
    def __init__(self, *args, **kwargs):
        super(_SnpSubset, self).__init__(*args, **kwargs)



import os
import numpy as np
import logging
import doctest
import unittest

class TestUtilTools(unittest.TestCase):

    def test_util_mapreduce1_testmod(self):
        import pysnptools.util.mapreduce1
        import pysnptools.util.pheno
        import pysnptools.util.mapreduce1.examples
        old_dir = os.getcwd()
        os.chdir(os.path.dirname(os.path.realpath(__file__))+"/mapreduce1")
        for mod in [
                    pysnptools.util,
                    pysnptools.util.pheno,
                    pysnptools.util.mapreduce1.examples,
                    pysnptools.util.mapreduce1.mapreduce,
                    pysnptools.util.mapreduce1.runner.local,
                    pysnptools.util.mapreduce1.runner.localinparts,
                    pysnptools.util.mapreduce1.runner.localmultiproc,
                    pysnptools.util.mapreduce1.runner.localmultithread,
                    ]:
            result = doctest.testmod(mod,optionflags=doctest.ELLIPSIS|doctest.NORMALIZE_WHITESPACE)
            assert result.failed == 0, "failed doc test: " + __file__
        os.chdir(old_dir)

    def test_local(self):
        from pysnptools.util.mapreduce1 import map_reduce
        from pysnptools.util.mapreduce1.runner import Local
        def holder1(n,runner):
            def mapper1(x):
                return x*x
            def reducer1(sequence):
               return sum(sequence)
            return map_reduce(range(n),mapper=mapper1,reducer=reducer1,runner=runner)
        assert 328350 == holder1(100,Local())
        
    def test_localinparts(self):
        from pysnptools.util.mapreduce1 import map_reduce
        from pysnptools.util.mapreduce1.runner import LocalInParts
        def holder1(n,runner):
            def mapper1(x):
                return x*x
            def reducer1(sequence):
               return sum(sequence)
            return map_reduce(range(n),mapper=mapper1,reducer=reducer1,runner=runner)
        holder1(100,LocalInParts(0,4)) #Run part 0 of 4 and save partial results to disk as '0.4.p'.
        holder1(100,LocalInParts(1,4)) #Run part 1 of 4 and save partial results to disk as '1.4.p'.
        holder1(100,LocalInParts(2,4)) #Run part 2 of 4 and save partial results to disk as '2.4.p'.
        holder1(100,LocalInParts(3,4)) #Run part 3 of 4 and save partial results to disk as '3.4.p'.
        assert 328350 == holder1(100,LocalInParts(4,4)) #Read the all partial results and then apply the reduce function and return the result.

    def test_localmultiproc(self):
        from pysnptools.util.mapreduce1 import map_reduce
        from pysnptools.util.mapreduce1.runner import LocalMultiProc
        def holder1(n,runner):
            def mapper1(x):
                return x*x
            def reducer1(sequence):
               return sum(sequence)
            return map_reduce(range(n),mapper=mapper1,reducer=reducer1,runner=runner)
        assert 328350 == holder1(100,LocalMultiProc(4))

    def test_localmultithread(self):
        from pysnptools.util.mapreduce1 import map_reduce
        from pysnptools.util.mapreduce1.runner import LocalMultiThread
        def holder1(n,runner):
            def mapper1(x):
                return x*x
            def reducer1(sequence):
               return sum(sequence)
            return map_reduce(range(n),mapper=mapper1,reducer=reducer1,runner=runner)
        assert 328350 == holder1(100,LocalMultiThread(4))
        
    def test_localinparts_with_weights(self):
        from pysnptools.util.mapreduce1 import map_reduce
        from pysnptools.util.mapreduce1.runner import LocalInParts
        def holder1(n,runner):
            def mapper1(x):
                return int(os.environ['TEST_ENVIRON'])
            def reducer1(sequence):
               return sum(sequence)+int(os.environ['TEST_ENVIRON'])
            return map_reduce(range(n),mapper=mapper1,reducer=reducer1,runner=runner)
        weights = [1,97,1,1]
        for i in range(5):
            environ = {'TEST_ENVIRON':str(i)}
            assert 'TEST_ENVIRON' not in os.environ
            result = holder1(100,LocalInParts(i,4,weights=weights,environ=environ)) #ignore all but the last result
            assert 'TEST_ENVIRON' not in os.environ
        assert result == 0*1+1*97+2*1+3*1+4

    def test_localinmultiproc_with_weights(self):
        from pysnptools.util.mapreduce1 import map_reduce
        from pysnptools.util.mapreduce1.runner import LocalMultiProc
        def holder1(n,runner):
            def mapper1(x):
                return int(os.environ['TEST_ENVIRON'])
            def reducer1(sequence):
               return sum(sequence)+int(os.environ['TEST_ENVIRON'])
            return map_reduce(range(n),mapper=mapper1,reducer=reducer1,runner=runner)
        weights = [1,97,1,1]
        def taskindex_to_environ(taskindex):
            return {'TEST_ENVIRON': str(taskindex)}
        runner = LocalMultiProc(4,weights=weights,taskindex_to_environ=taskindex_to_environ)
        assert 'TEST_ENVIRON' not in os.environ
        result = holder1(100,runner)
        assert result == 0*1+1*97+2*1+3*1+4

    def test_sub_matrix(self):
        import pysnptools.util as pstutil
        np.random.seed(0) # set seed so that results are deterministic
        matrix = np.random.rand(12,7) # create a 12 x 7 ndarray
        submatrix = pstutil.sub_matrix(matrix,[0,2,11],[6,5,4,3,2,1,0])
        assert  matrix[2,0] == submatrix[1,6] #The row # 2 is now #1, the column #0 is now #6.

        np.random.seed(0) # set seed so that results are deterministic
        matrix = np.random.rand(12,7,3) # create a 12 x 7 ndarray
        submatrix = pstutil.sub_matrix(matrix,[0,2,11],[6,5,4,3,2,1,0])
        assert  matrix[2,0,1] == submatrix[1,6,1] #The row # 2 is now #1, the column #0 is now #6.



def getTestSuite():
    """
    set up composite test suite
    """

    test_suite = unittest.TestSuite([])
    test_suite.addTests(unittest.TestLoader().loadTestsFromTestCase(TestUtilTools))
    return test_suite

if __name__ == '__main__':
    logging.basicConfig(level=logging.INFO)

    suites = getTestSuite()
    r = unittest.TextTestRunner(failfast=False)
    ret = r.run(suites)
    assert ret.wasSuccessful()



'''
Runs a distributable job locally in one process. Returns the value of the job.

See SamplePi.py for examples.
'''

from unittest.mock import patch
from pysnptools.util.mapreduce1.runner import Runner, _run_all_in_memory
import os, sys
import logging

class Local(Runner):
    '''
    A :class:`.Runner` that runs a :func:`.map_reduce` locally. To save memory, it will feed the results of the mapper to the reducer as those results are computed.

    **Constructor:**
        :Parameters: * **mkl_num_threads** (*number*) -- (default None) Limit on the number threads used by the NumPy MKL library.
        :Parameters: * **logging_handler** (*stream*) --  (default stdout) Where to output logging messages.
        
        :Example:

        >>> from pysnptools.util.mapreduce1 import map_reduce
        >>> from pysnptools.util.mapreduce1.runner import Local
        >>> def holder1(n,runner):
        ...     def mapper1(x):
        ...         return x*x
        ...     def reducer1(sequence):
        ...        return sum(sequence)
        ...     return map_reduce(range(n),mapper=mapper1,reducer=reducer1,runner=runner)
        >>> holder1(100,Local())
        328350

    '''
    def __init__(self, mkl_num_threads = None, logging_handler=logging.StreamHandler(sys.stdout)):
        logger = logging.getLogger()
        if not logger.handlers:
            logger.setLevel(logging.INFO)
        for h in list(logger.handlers):
            logger.removeHandler(h)
        logger.addHandler(logging_handler)
        if logger.level == logging.NOTSET:
            logger.setLevel(logging.INFO)
        self.mkl_num_threads = mkl_num_threads

    def run(self, distributable):
        _JustCheckExists().input(distributable)
        with patch.dict('os.environ', {'MKL_NUM_THREADS': str(self.mkl_num_threads)} if self.mkl_num_threads is not None else {}) as _:
            result = _run_all_in_memory(distributable)
        _JustCheckExists().output(distributable)
        return result

class _JustCheckExists(object): #Implements ICopier

    def __init__(self,doPrintOutputNames=False):
        self.doPrintOutputNames = doPrintOutputNames
    
    def input(self,item):
        if isinstance(item, str):
            if not os.path.exists(item): raise Exception("Missing input file '{0}'".format(item))
        elif hasattr(item,"copyinputs"):
            item.copyinputs(self)
        # else -- do nothing

    def output(self,item):
        if isinstance(item, str):
            if not os.path.exists(item): raise Exception("Missing output file '{0}'".format(item))
            if self.doPrintOutputNames:
                print(item)
        elif hasattr(item,"copyoutputs"):
            item.copyoutputs(self)
        # else -- do nothing

if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO)

    import doctest
    doctest.testmod(optionflags=doctest.ELLIPSIS)



from io import StringIO
import sys
import unittest
import logging
import os
import tempfile
import shutil
import time
from pysnptools.util.filecache import LocalCache, PeerToPeer, Hashdown


class TestFileCache(unittest.TestCase):
    def test_local_file(self):
        logging.info("test_local_file")

        temp_dir = self._temp_dir()

        def storage_closure():
            return LocalCache(temp_dir)

        self._write_and_read(storage_closure())
        self._distribute(storage_closure)

    def test_hashdown(self):
        logging.info("test_hashdown")

        from pysnptools.util.filecache.test import TestFileCache as self

        logging.basicConfig(level=logging.INFO)
        logging.info("test_hashdown")

        url = "https://github.com/fastlmm/PySnpTools/raw/9de8e93a91b330b064b482c918a38104904b45c0/pysnptools"
        hashdown0 = Hashdown(url, allow_unknown_files=True)
        for file in [
            "examples/toydata.5chrom.bed",
            "examples/toydata.bim",
            "examples/toydata.fam",
            "util/util.py",
        ]:
            hashdown0.file_exists(file)
        _ = hashdown0.file_to_hash
        hashdown = Hashdown(
            url, file_to_hash=hashdown0.file_to_hash, allow_unknown_files=False
        )

        # Clear the directory
        assert self._is_error(
            lambda: hashdown.rmtree()
        )  # raise error because this is read-only
        # Rule: After you clear a directory, nothing is in it
        assert 4 == self._len(
            hashdown.walk()
        )  # returns the files in the file_to_hash_file
        assert not hashdown.file_exists("test.txt")
        assert not hashdown.file_exists("main.txt/test.txt")
        assert not hashdown.file_exists(r"main.txt\test.txt")
        assert self._is_error(
            lambda: hashdown.file_exists("test.txt/")
        )  # Can't query something that can't be a file name
        assert self._is_error(
            lambda: hashdown.file_exists("../test.txt")
        )  # Can't leave the current directory
        if os.name == "nt":
            assert self._is_error(
                lambda: hashdown.file_exists(r"c:\test.txt")
            )  # Can't leave the current directory

        # Rule: '/' and '\' are both OK, but you can't use ':' or '..' to leave the current root.
        assert self._is_error(lambda: 0 == self._len(hashdown.walk("..")))
        assert 0 == self._len(hashdown.walk("..x"))
        assert 0 == self._len(
            hashdown.walk("test.txt")
        )  # This is ok, because test.txt doesn't exist and therefore isn't a file
        assert 0 == self._len(hashdown.walk("a/b"))
        assert 0 == self._len(hashdown.walk("a\\b"))  # Backslash or forward is fine
        assert self._is_error(lambda: len(hashdown.walk("/")))  # Can't start with '/'
        assert self._is_error(lambda: len(hashdown.walk(r"\\")))  # Can't start with '\'
        assert self._is_error(
            lambda: len(hashdown.walk(r"\\computer1\share\3"))
        )  # Can't start with UNC

        # It should be there and be a file
        assert hashdown.file_exists("examples/toydata.bim")
        file_list = list(hashdown.walk())
        assert len(file_list) == 4 and "examples/toydata.bim" in file_list
        file_list2 = list(hashdown.walk("examples"))
        assert len(file_list2) == 3 and "examples/toydata.bim" in file_list2
        assert self._is_error(
            lambda: hashdown.join("examples/toydata.bim")
        )  # Can't create a directory where a file exists
        assert self._is_error(
            lambda: list(hashdown.walk("examples/toydata.bim"))
        )  # Can't have directory where a file exists

        # Read it
        assert (
            hashdown.load("examples/toydata.bim").split("\n")[0]
            == "1\tnull_0\t0\t1\tL\tH"
        )
        assert hashdown.file_exists("examples/toydata.bim")
        assert self._is_error(
            lambda: hashdown.load("examples")
        )  # This is an error because examples is actually a directory and they can't be opened for reading

        # Can query modified time of file.
        assert self._is_error(
            lambda: hashdown.getmtime("a/b/c.txt")
        ), "Can't get mod time from file that doesn't exist"
        assert self._is_error(
            lambda: hashdown.getmtime("examples")
        ), "Can't get mod time from directory"
        assert (
            hashdown.getmtime("examples/toydata.bim") == 0.0
        ), "expect all mod times to be 0"

        # try to write
        assert self._is_error(
            lambda: hashdown.save("main.txt", "")
        ), "Can't write. It's read only"
        # try to remove
        assert self._is_error(
            lambda: hashdown.remove("examples/toydata.bim")
        ), "Can't remove. It's read only"
        assert self._is_error(lambda: hashdown.rmtree()), "Can't remove. It's read only"
        # what if files are not local?
        shutil.rmtree(hashdown.directory, ignore_errors=True)
        # It should be there and be a file
        assert hashdown.file_exists("examples/toydata.bim")
        file_list = list(hashdown.walk())
        assert len(file_list) == 4 and "examples/toydata.bim" in file_list
        file_list2 = list(hashdown.walk("examples"))
        assert len(file_list2) == 3 and "examples/toydata.bim" in file_list2
        assert (
            hashdown.load("examples/toydata.bim").split("\n")[0]
            == "1\tnull_0\t0\t1\tL\tH"
        )
        # What if put bad file locally?
        os.rename(
            hashdown.directory + "/examples/toydata.bim",
            hashdown.directory + "/examples/toydata.fam",
        )
        assert (
            hashdown.load("examples/toydata.fam").split("\n")[0] == "per0 per0 0 0 0 0"
        )
        # What if hash doesn't match web hash?
        hashdown.file_to_hash["examples/toydata.fam"] = "WRONG"
        assert self._is_error(
            lambda: hashdown.load("examples/toydata.fam")
        ), "unexpected hash"
        # writing out file_to_hash and read_file_to_hash

        hashdown.save_hashdown("ignore/toydata.hashdown.json")
        hashdown2 = Hashdown.load_hashdown("ignore/toydata.hashdown.json")
        # It should be there and be a file
        assert hashdown2.file_exists("examples/toydata.bim")
        file_list = list(hashdown2.walk())
        assert len(file_list) == 4 and "examples/toydata.bim" in file_list
        file_list2 = list(hashdown2.walk("examples"))
        assert len(file_list2) == 3 and "examples/toydata.bim" in file_list2
        assert (
            hashdown2.load("examples/toydata.bim").split("\n")[0]
            == "1\tnull_0\t0\t1\tL\tH"
        )

    def test_peer_to_peer(self):
        logging.info("test_peer_to_peer")

        temp_dir = self._temp_dir()

        def storage_closure():
            def id_and_path_function():
                self.count += 1
                return self.count, temp_dir + "/{0}".format(self.count)

            storage = PeerToPeer(
                common_directory=temp_dir + "/common",
                id_and_path_function=id_and_path_function,
            )
            return storage

        self._write_and_read(storage_closure())
        self._distribute(storage_closure)

    @staticmethod
    def file_name(self, testcase_name):
        temp_fn = os.path.join(self.tempout_dir, testcase_name + ".txt")
        if os.path.exists(temp_fn):
            os.remove(temp_fn)
        return temp_fn

    tempout_dir = "tempout/one_milp2p"

    @classmethod
    def setUpClass(self):
        self.temp_parent = os.path.join(tempfile.gettempdir(), format(hash(os.times())))
        self.count = 0

    def _temp_dir(self):
        self.count += 1
        return "{0}/{1}".format(self.temp_parent, self.count)

    @classmethod
    def tearDownClass(self):
        not os.path.exists(self.temp_parent) or shutil.rmtree(self.temp_parent)

    @staticmethod
    def _is_error(lambda0):
        try:
            lambda0()
        except Exception as e:
            logging.debug(str(e))
            return True
        return False

    @staticmethod
    def _len(sequence):
        len = 0
        for item in sequence:
            len += 1
        return len

    def _write_and_read(self, storage):
        test_storage = storage.join("test_snps")
        # !!!How to induce an error: create a 'test_snps' file at the top level then try to create
        # an empty directory with the same name

        # Clear the directory
        test_storage.rmtree()
        # Rule: After you clear a directory, nothing is in it
        assert 0 == self._len(test_storage.walk())
        assert not test_storage.file_exists("test.txt")
        assert not test_storage.file_exists("main.txt/test.txt")
        assert not test_storage.file_exists(r"main.txt\test.txt")
        assert self._is_error(
            lambda: test_storage.file_exists("test.txt/")
        )  # Can't query something that can't be a file name
        assert self._is_error(
            lambda: test_storage.file_exists("../test.txt")
        )  # Can't leave the current directory
        if os.name == "nt":
            assert self._is_error(
                lambda: test_storage.file_exists(r"c:\test.txt")
            )  # Can't leave the current directory

        # Rule: '/' and '\' are both OK, but you can't use ':' or '..' to leave the current root.
        assert 0 == self._len(test_storage.walk())
        assert self._is_error(lambda: 0 == self._len(test_storage.walk("..")))
        assert 0 == self._len(test_storage.walk("..x"))
        assert 0 == self._len(
            test_storage.walk("test.txt")
        )  # This is ok, because test.txt doesn't exist and therefore isn't a file
        assert 0 == self._len(test_storage.walk("a/b"))
        assert 0 == self._len(test_storage.walk("a\\b"))  # Backslash or forward is fine
        assert self._is_error(
            lambda: len(test_storage.walk("/"))
        )  # Can't start with '/'
        assert self._is_error(
            lambda: len(test_storage.walk(r"\\"))
        )  # Can't start with '\'
        assert self._is_error(
            lambda: len(test_storage.walk(r"\\computer1\share\3"))
        )  # Can't start with UNC

        # Clear the directory, again
        test_storage.rmtree()
        assert 0 == self._len(test_storage.walk())
        test_storage.rmtree("main.txt")
        assert 0 == self._len(test_storage.walk("main.txt"))
        assert 0 == self._len(test_storage.walk())

        # Write to it.
        assert self._is_error(lambda: test_storage.save("../test.txt"))
        test_storage.save("main.txt/test.txt", "test\n")
        # Rule: It's an error to write to a file or directory that already exists
        assert self._is_error(lambda: test_storage.save("main.txt"))
        assert self._is_error(lambda: test_storage.save("main.txt/test.txt"))
        assert self._is_error(
            lambda: list(test_storage.walk("main.txt/test.txt"))
        ), "Rule: It's an error to walk a file (but recall that it's OK to walk a folder that doesn't exist)"

        # It should be there and be a file
        assert test_storage.file_exists("main.txt/test.txt")
        file_list = list(test_storage.walk())
        assert len(file_list) == 1 and file_list[0] == "main.txt/test.txt"
        file_list2 = list(test_storage.walk("main.txt"))
        assert len(file_list2) == 1 and file_list2[0] == "main.txt/test.txt"
        assert self._is_error(
            lambda: test_storage.join("main.txt/test.txt")
        )  # Can't create a directory where a file exists
        assert self._is_error(
            lambda: list(test_storage.walk("main.txt/test.txt"))
        )  # Can't create a directory where a file exists
        assert self._is_error(
            lambda: test_storage.rmtree("main.txt/test.txt")
        )  # Can't delete a directory where a file exists

        # Read it
        assert test_storage.load("main.txt/test.txt") == "test\n"
        assert test_storage.file_exists("main.txt/test.txt")
        assert self._is_error(
            lambda: test_storage.load("main.txt")
        )  # This is an error because main.txt is actually a directory and they can't be opened for reading

        # Remove it
        test_storage.remove("main.txt/test.txt")
        assert self._is_error(
            lambda: test_storage.remove("main.txt/test.txt")
        )  # Can't delete a file that doesn't exist
        assert not test_storage.file_exists("main.txt/test.txt")
        assert 0 == self._len(test_storage.walk())
        assert 0 == self._len(test_storage.walk("main.txt"))
        assert 0 == self._len(test_storage.walk("main.txt/test.txt"))  # Now allowed.

        #  writing zero length files is OK
        #  File share has a special file called "main.txt". Can we mess things up by using 'main.txt' as a directory
        #  name, too. It's OK to write to a file in a directory that used to exist, but now has no files.
        test_storage.save("main.txt", "")
        assert test_storage.file_exists("main.txt")
        file_list = list(test_storage.walk())
        assert len(file_list) == 1 and file_list[0] == "main.txt"
        assert test_storage.load("main.txt") == ""
        assert test_storage.file_exists("main.txt")

        # Can query modified time of file. It will be later, later.
        assert self._is_error(
            lambda: test_storage.getmtime("a/b/c.txt")
        ), "Can't get mod time from file that doesn't exist"
        test_storage.save("a/b/c.txt", "")
        m1 = test_storage.getmtime("a/b/c.txt")
        assert self._is_error(
            lambda: test_storage.getmtime("a/b")
        ), "Can't get mod time from directory"
        assert (
            test_storage.getmtime("a/b/c.txt") == m1
        ), "expect mod time to stay the same"
        assert test_storage.load("a/b/c.txt") == ""
        assert (
            test_storage.getmtime("a/b/c.txt") == m1
        ), "reading a file doesn't change its mod time"
        test_storage.remove("a/b/c.txt")
        assert self._is_error(
            lambda: test_storage.getmtime("a/b/c.txt")
        ), "Can't get mod time from file that doesn't exist"
        time.sleep(1)  # Sleep one second
        test_storage.save("a/b/c.txt", "")
        test_storage.save("a/b/d.txt", "")
        assert (
            test_storage.getmtime("a/b/c.txt") > m1
        ), "A file created later (after a pause) will have a later mod time"
        assert (
            test_storage.getmtime("a/b/d.txt") > m1
        ), "A file created later (after a pause) will have a later mod time"
        assert test_storage.getmtime("a/b/d.txt") >= test_storage.getmtime(
            "a/b/c.txt"
        ), "A file created later (with no pause) will have a later or equal mod time"

    # Look for code that isn't covered and make test cases for it
    # Distributed testing
    # What happens if you don't close an open_write? Undefined

    def _distribute(self, storage_lambda):
        storage1 = storage_lambda()
        storage2 = storage_lambda()
        # clear everything on #1
        storage1.rmtree()
        # 1 and #2 agree nothing is there
        assert 0 == self._len(storage1.walk())
        assert 0 == self._len(storage2.walk())
        # write on #1
        storage1.save("a/b/c.txt", "Hello")
        # read on #2
        assert storage2.load("a/b/c.txt") == "Hello"
        # read on #2, again and manually see that it caches
        assert storage2.load("a/b/c.txt") == "Hello"
        # remove on #1
        storage1.remove("a/b/c.txt")
        # assert not exists on #2
        assert not storage2.file_exists("a/b/c.txt")
        # write on #1
        storage1.save("a/b/c.txt", "Hello")
        # read on #2
        assert storage2.load("a/b/c.txt") == "Hello"
        # remove on #1
        storage1.remove("a/b/c.txt")
        # write something different on #1
        storage1.save("a/b/c.txt", "There")
        # read on #2 and see that it is different.
        assert storage2.load("a/b/c.txt") == "There"

    def test_util_filecache_testmod(self):
        import doctest
        import pysnptools.util.filecache

        old_dir = os.getcwd()
        os.chdir(os.path.dirname(os.path.realpath(__file__)))
        for mod in [
            pysnptools.util.filecache,
            pysnptools.util.filecache.filecache,
            pysnptools.util.filecache.localcache,
            pysnptools.util.filecache.peertopeer,
            pysnptools.util.filecache.hashdown,
        ]:
            buffer = StringIO()
            old_stdout = sys.stdout
            sys.stdout = buffer  # Redirect the stdout to the buffer

            try:
                result = doctest.testmod(
                    mod, optionflags=doctest.ELLIPSIS | doctest.NORMALIZE_WHITESPACE
                )
                if result.failed:
                    assert (
                        result.failed == 0
                    ), f"Doc test failed in {mod.__name__} with the following output:\n{buffer.getvalue()}"
            finally:
                sys.stdout = old_stdout
                buffer.close()
        os.chdir(old_dir)


def getTestSuite():
    """
    set up composite test suite
    """

    test_suite = unittest.TestSuite([])
    test_suite.addTests(unittest.TestLoader().loadTestsFromTestCase(TestFileCache))
    return test_suite


if __name__ == "__main__":
    logging.basicConfig(level=logging.WARN)
    suites = getTestSuite()

    r = unittest.TextTestRunner(failfast=True)  # !!! should be false
    ret = r.run(suites)
    assert ret.wasSuccessful()

    logging.info("done")



import numpy as np
from pysnptools.pstreader import PstNpz
from pysnptools.kernelreader import KernelReader
import logging


class KernelNpz(KernelReader,PstNpz):
    '''
    A :class:`.KernelReader` for reading \*.kernel.npz files from disk.

    See :class:`.KernelReader` for general examples of using KernelReaders.

    The general NPZ format is described `here <http://docs.scipy.org/doc/numpy/reference/generated/numpy.savez.html>`__. The KernelNpz format stores
    val, iid0, and iid1 information in NPZ format.
   
    **Constructor:**
        :Parameters: * **filename** (*string*) -- The KernelNpz file to read.

        :Example:

        >>> from pysnptools.kernelreader import KernelNpz
        >>> from pysnptools.util import example_file # Download and return local file name
        >>> npz_file = example_file('pysnptools/examples/toydata.kernel.npz')
        >>> data_on_disk = KernelNpz(npz_file)
        >>> print(data_on_disk.iid_count)
        500

    **Methods beyond** :class:`.KernelReader`

    '''

    def __init__(self,*args, **kwargs):
        super(KernelNpz, self).__init__(*args, **kwargs)

    @property
    def row(self):
        self._run_once()
        if self._row.dtype.type is not np.str_:
            if self._row is not self._col:
                self._row = np.array(self._row,dtype='str')
            else:
                self._row = np.array(self._row,dtype='str')
                self._col = self._row
        return self._row

    @property
    def col(self):
        self._run_once()
        if self._col.dtype.type is not np.str_:
            if self._row is not self._col:
                self._col = np.array(self._col,dtype='str')
            else:
                self._row = np.array(self._row,dtype='str')
                self._col = self._row
        return self._col


    @staticmethod
    def write(filename, kerneldata):
        """Writes a :class:`KernelData` to KernelNpz format and returns the :class:`.KernelNpz`.

        :param filename: the name of the file to create
        :type filename: string
        :param kerneldata: The in-memory data that should be written to disk.
        :type kerneldata: :class:`KernelData`
        :rtype: :class:`.KernelNpz`

        >>> from pysnptools.snpreader import Bed
        >>> from pysnptools.standardizer import Unit
        >>> import pysnptools.util as pstutil
        >>> from pysnptools.util import example_file # Download and return local file name
        >>> bed_file = example_file('pysnptools/examples/toydata.5chrom.*','*.bed')
        >>> kerneldata = Bed(bed_file,count_A1=False).read_kernel(Unit())     # Create a kernel from the data in the Bed file
        >>> pstutil.create_directory_if_necessary("tempdir/toydata.kernel.npz")
        >>> KernelNpz.write("tempdir/toydata.kernel.npz",kerneldata)      # Write data in KernelNpz format
        KernelNpz('tempdir/toydata.kernel.npz')
        """
        PstNpz.write(filename,kerneldata)
        return KernelNpz(filename)

if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO)

    import doctest
    doctest.testmod(optionflags=doctest.ELLIPSIS)



from pysnptools.pstreader import PstNpz
from pysnptools.distreader import DistReader
import logging
import scipy as np
import warnings


class DistNpz(PstNpz, DistReader):
    """
    A :class:`.DistReader` for reading \*.dist.npz files from disk.

    See :class:`.DistReader` for general examples of using DistReaders.

    The general NPZ format is described `here <http://docs.scipy.org/doc/numpy/reference/generated/numpy.savez.html>`__. The DistNpz format stores
    val, iid, sid, and pos information in NPZ format.

    **Constructor:**
        :Parameters: * **filename** (*string*) -- The DistNpz file to read.

        :Example:

        >>> from pysnptools.distreader import DistNpz
        >>> from pysnptools.util import example_file # Download and return local file name
        >>> npz_file = example_file("pysnptools/examples/toydata10.dist.npz")
        >>> data_on_disk = DistNpz(npz_file)
        >>> print((data_on_disk.iid_count, data_on_disk.sid_count))
        (25, 10)

    **Methods beyond** :class:`.DistReader`

    """

    def __init__(self, *args, **kwargs):
        super(DistNpz, self).__init__(*args, **kwargs)

    @property
    def row(self):
        self._run_once()
        if self._row.dtype.type is not np.str_:
            self._row = np.array(self._row, dtype="str")
        return self._row

    @property
    def col(self):
        self._run_once()
        if self._col.dtype.type is not np.str_:
            self._col = np.array(self._col, dtype="str")
        return self._col

    @staticmethod
    def write(filename, distdata):
        """Writes a :class:`DistData` to DistNpz format and returns the :class:`.DistNpz`

        :param filename: the name of the file to create
        :type filename: string
        :param distdata: The in-memory data that should be written to disk.
        :type distdata: :class:`DistData`
        :rtype: :class:`.DistNpz`

        >>> from pysnptools.distreader import DistNpz, DistHdf5
        >>> import pysnptools.util as pstutil
        >>> from pysnptools.util import example_file # Download and return local file name
        >>> hdf5_file = example_file("pysnptools/examples/toydata.iidmajor.dist.hdf5")
        >>> distdata = DistHdf5(hdf5_file)[:,:10].read()     # Read first 10 snps from DistHdf5 format
        >>> pstutil.create_directory_if_necessary("tempdir/toydata10.dist.npz")
        >>> DistNpz.write("tempdir/toydata10.dist.npz",distdata)          # Write data in DistNpz format
        DistNpz('tempdir/toydata10.dist.npz')
        """
        row_ascii = np.array(
            distdata.row, dtype="S"
        )  # !!! would be nice to avoid this copy when not needed.
        col_ascii = np.array(
            distdata.col, dtype="S"
        )  # !!! would be nice to avoid this copy when not needed.
        np.savez(
            filename,
            row=row_ascii,
            col=col_ascii,
            row_property=distdata.row_property,
            col_property=distdata.col_property,
            val=distdata.val,
        )
        logging.debug("Done writing " + filename)
        return DistNpz(filename)


if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO)

    import doctest

    doctest.testmod()



import numpy as np
from pysnptools.standardizer import Standardizer
import logging
import warnings


class DiagKtoN(Standardizer):
    """
    Both a :class:`.Standardizer` and a A :class:`.KernelStandardizer`.

    When applied to a :class:`.SnpData`, it multiplies the SNP values by the square root of a factor such that a kernel
    constructed from the SNP values will have a diagonal that sums to iid_count. This class thus
    standardizes the kernel before it is even constructed.

    When applied to a :class:`.KernelData`, it multiplies the kernel values by the a factor such that a kernel
    will have a diagonal that sums to iid_count.

    See :class:`.Standardizer` for more information about standardization.

    **Constructor:**
        :Parameters: * **deprecated_iid_count** (*int*) (optional) -- Deprecated.


    Example of DiagKtoN to :class:`.SnpData`:

    >>> from pysnptools.standardizer import DiagKtoN, Unit, Identity
    >>> from pysnptools.snpreader import Bed
    >>> from pysnptools.util import example_file # Download and return local file name
    >>> bedfile = example_file("tests/datasets/all_chr.maf0.001.N300.*","*.bed")
    >>> snpdata1 = Bed(bedfile,count_A1=False).read().standardize(Unit()).standardize(DiagKtoN())
    >>> kernel1 = snpdata1.read_kernel(Identity())
    >>> print('{0:.6f}'.format(np.diag(kernel1.val).sum()))
    300.000000

    Example of DiagKtoN to :class:`.KernelData`:

    >>> import numpy as np
    >>> from pysnptools.standardizer import DiagKtoN, Unit, Identity
    >>> from pysnptools.snpreader import Bed
    >>> from pysnptools.util import example_file # Download and return local file name
    >>> bedfile = example_file("tests/datasets/all_chr.maf0.001.N300.*","*.bed")
    >>> snpdata1 = Bed(bedfile,count_A1=False).read().standardize(Unit())
    >>> kernel1 = snpdata1.read_kernel(DiagKtoN(),block_size=None)
    >>> print('{0:.6f}'.format(np.diag(kernel1.val).sum()))
    300.000000

    """

    """diag(K)=N standardization of the data"""

    def __init__(self, deprecated_iid_count=None):
        super(DiagKtoN, self).__init__()
        if deprecated_iid_count is not None:
            warnings.warn(
                "'iid_count' is deprecated (and not needed, since can get iid_count from SNPs val's first dimension",
                DeprecationWarning,
            )

    def _standardize_kernel(
        self,
        kerneldata,
        return_trained=False,
        force_python_only=False,
        num_threads=None,
    ):
        factor = float(kerneldata.iid_count) / np.diag(kerneldata.val).sum()

        if abs(factor - 1.0) > 1e-15:
            kerneldata._val *= factor

        if return_trained:
            return kerneldata, DiagKtoNTrained(factor)
        else:
            return kerneldata

    def standardize(
        self,
        input,
        block_size=None,
        return_trained=False,
        force_python_only=False,
        num_threads=None,
    ):
        from pysnptools.kernelreader import KernelReader

        if block_size is not None:
            warnings.warn(
                "block_size is deprecated (and not needed, since standardization is in-place",
                DeprecationWarning,
            )
        if isinstance(input, KernelReader) and hasattr(input, "val"):
            return self._standardize_kernel(
                input,
                return_trained=return_trained,
                force_python_only=force_python_only,
                num_threads=num_threads,
            )
        else:
            return self._standardize_snps(
                input,
                return_trained=return_trained,
                force_python_only=force_python_only,
                num_threads=num_threads,
            )

    def _standardize_snps(
        self, snps, return_trained=False, force_python_only=False, num_threads=None
    ):
        if hasattr(snps, "val"):
            val = snps.val
        else:
            warnings.warn(
                "standardizing an nparray instead of a SnpData is deprecated",
                DeprecationWarning,
            )
            val = snps

        vec = val.reshape(-1, order="A")
        # make sure no copy was made
        assert not vec.flags["OWNDATA"]
        squared_sum = vec.dot(vec)
        factor = float(val.shape[0]) / squared_sum

        if abs(factor - 1.0) > 1e-15:
            val *= np.sqrt(factor)

        if return_trained:
            return snps, DiagKtoNTrained(factor)
        else:
            return snps

    def __repr__(self):
        return "{0}()".format(self.__class__.__name__)


# !!!move to its own file
class DiagKtoNTrained(Standardizer):
    """Both a :class:`.Standardizer` and a A :class:`.KernelStandardizer`.

    When applied to a :class:`.SnpData`, DiagKtoN standardizes one set of SNP data based on another set of SNP data.

    When applied to a :class:`.KernelData`, DiagKtoN standardizes one kernel data based on another kernel data.

    See :class:`.Standardizer` for more information about standardization.

    **Constructor:**
        :Parameters: * **factor** (*float*) -- The number what when multiplied into the kernel values will make the diagonal of the kernel values to  sum to iid_count.

    """

    def __init__(self, factor):
        super(DiagKtoNTrained, self).__init__()
        self.factor = factor

    def standardize(
        self,
        input,
        block_size=None,
        return_trained=False,
        force_python_only=False,
        num_threads=None,
    ):
        if block_size is not None:
            warnings.warn(
                "block_size is deprecated (and not needed, since standardization is in-place",
                DeprecationWarning,
            )

        from pysnptools.kernelreader import KernelReader

        if isinstance(input, KernelReader) and hasattr(input, "val"):
            return self._standardize_kernel(
                input,
                return_trained=return_trained,
                force_python_only=force_python_only,
                num_threads=num_threads,
            )
        else:
            return self._standardize_snps(
                input,
                return_trained=return_trained,
                force_python_only=force_python_only,
                num_threads=num_threads,
            )  # LATER test coverage

    @property
    def is_constant(self):
        return abs(self.factor - 1.0) < 1e-15

    def _standardize_snps(
        self, snps, return_trained=False, force_python_only=False, num_threads=None
    ):
        if hasattr(snps, "val"):
            val = snps.val  # LATER test coverage
        else:
            warnings.warn(
                "standardizing an nparray instead of a SnpData is deprecated",
                DeprecationWarning,
            )  # LATER test coverage
            val = snps

        if not self.is_constant:
            val *= np.sqrt(self.factor)

        if return_trained:
            return snps, self
        else:
            return snps

    def _standardize_kernel(
        self,
        kerneldata,
        return_trained=False,
        force_python_only=False,
        num_threads=None,
    ):
        if not self.is_constant:
            kerneldata._val *= self.factor

        if return_trained:
            return kerneldata, self
        else:
            return kerneldata

    def __repr__(self):
        return "{0}({1})".format(self.__class__.__name__, self.factor)


if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO)

    import doctest

    doctest.testmod(optionflags=doctest.ELLIPSIS)



import logging
from pysnptools.util.mapreduce1.runner import *
import os
import cPickle as pickle
import subprocess, sys, os.path
import multiprocessing
import pysnptools.util as pstutil
import pdb
from collections import defaultdict
import tarfile as tarfileLibrary
import ctypes
import datetime
from tempfile import TemporaryFile


class Hadoop(Runner):

    """
    Old code to run on Hadoop. Not currently supported.
    """

    fileshare = "/user"

    def __init__(
        self,
        taskcount,
        mapmemory=-1,
        reducememory=-1,
        mkl_num_threads=None,
        queue="default",
        skipsourcecheck=False,
        skipdatacheck=False,
        logging_handler=logging.StreamHandler(sys.stdout),
    ):
        logger = logging.getLogger()
        if not logger.handlers:
            logger.setLevel(logging.INFO)
        for h in list(logger.handlers):
            logger.removeHandler(h)
        logger.addHandler(logging_handler)
        if logger.level == logging.NOTSET:
            logger.setLevel(logging.INFO)

        logging.info("constructing Hadoop runner")
        self.taskcount = taskcount
        self.mapmemory = mapmemory
        self.reducememory = reducememory
        self.mkl_num_threads = mkl_num_threads
        self.queue = queue
        self.skipsourcecheck = skipsourcecheck
        self.skipdatacheck = skipdatacheck
        logging.info("done constructing Hadoop runner")

    def run(self, distributable):
        logging.info("Hadoop runner is running a distributable")

        # Check that the local machine has python path set
        localpythonpath = os.environ.get(
            "PYTHONPATH"
        )  # !!should it be able to work without pythonpath being set (e.g. if there was just one file)? Also, is None really the return or is it an exception.
        if localpythonpath == None:
            raise Exception("Expect local machine to have 'pythonpath' set")

        remotewd, run_dir_abs, run_dir_rel = self.create_run_dir()

        result_local = os.path.join(run_dir_rel, "result.p")
        result_hdfs = "hdfs:" + os.path.join(run_dir_abs, "result.p").replace("\\", "/")
        result_remote = os.path.join(distributable.tempdirectory, "result.p")

        fileInWorkingDirectoryList = []

        tgzListPythonSettings = self.FindOrCreatePythonSettings(remotewd)

        remotepythonpath, tgzListPythonPath = self.FindOrCreateRemotePythonPath(
            localpythonpath, remotewd
        )

        tgzList = []
        inputOutputCopier = HadoopCopier(
            remotewd, fileInWorkingDirectoryList, tgzList, self.skipdatacheck
        )  # Create the object that copies input and output files to where they are needed
        inputOutputCopier.input(
            distributable
        )  # copy of the input files to where they are needed (i.e. the cluster)

        batfilename_abs_list = self.create_bat_file(
            distributable,
            remotepythonpath,
            remotewd,
            run_dir_abs,
            run_dir_rel,
            result_remote,
            result_hdfs,
        )

        self.submit_to_cluster(
            batfilename_abs_list,
            fileInWorkingDirectoryList,
            tgzList,
            tgzListPythonPath,
            tgzListPythonSettings,
            distributable,
            remotewd,
            run_dir_abs,
            run_dir_rel,
        )

        inputOutputCopier.output(
            distributable
        )  # copy the output file from where they were created (i.e. the cluster) to the local computer

        subprocess.check_output(
            "%HADOOP_HOME%\\bin\\hadoop fs -copyToLocal {0} {1}\n".format(
                result_hdfs, result_local
            ),
            stderr=subprocess.STDOUT,
            shell=True,
        )
        with open(result_local, mode="rb") as f:
            result = pickle.load(f)

        # logging.info('Done: Hadoop runner is running a distributable. Returns {0}'.format(result))
        return result

    def submit_to_cluster(
        self,
        batfilename_rel_list,
        fileInWorkingDirectoryList,
        tgzList,
        tgzListPythonPath,
        tgzListPythonSettings,
        distributable,
        remotewd,
        run_dir_abs,
        run_dir_rel,
    ):
        logging.info("Hadoop runner is submitting to cluster")

        # !! e.g. hdfs://rr1-n13-02-c02/user/carlk/inputs.tgz#inputs,hdfs://rr1-n13-02-c02/user/carlk/datasets.tgz#datasets,hdfs://rr1-n13-02-c02/user/carlk/src.tgz#src
        # !! could do this functionally
        archivesStringList = []
        for tgz in tgzList:
            archiveString = "hdfs:{0}#{1}".format(tgz[1], os.path.splitext(tgz[0])[0])
            archivesStringList.append(archiveString)
        archivesStringList.append(tgzListPythonSettings)
        for tgz in tgzListPythonPath:
            archivesStringList.append(tgz)

        # e.g. distMapper.bat,distReducer.bat
        filesString = ",".join(batfilename_rel_list + fileInWorkingDirectoryList)

        taskIndexDir = run_dir_rel + os.path.sep + "input"
        pstutil.create_directory_if_necessary(taskIndexDir, isfile=False)

        # zgoal = int(SP.ceil(SP.log(self.taskcount)/SP.log(10)))
        with open(
            taskIndexDir + os.path.sep + "taskIndexList.txt", "w"
        ) as taskIndexListFile:
            for taskIndex in xrange(self.taskcount):
                taskIndexListFile.write(
                    "{0}\n".format(taskIndex)
                )  # str(taskIndex).zfill(zgoal)))

        # hadoop fs -rmr runs/2013-08-02_13_51_42
        # hadoop fs -copyFromLocal runs\2013-08-02_13_51_42 runs/2013-08-02_13_51_42
        # hadoop jar %HADOOP_HOME%\lib\hadoop-streaming.jar ^
        #        -archives "hdfs:/user/carlk/source/carlkextranet05312013/ERG01/src/tests/datasets.2013-07-31_11_12_11.tgz#datasets,hdfs:/user/carlk/runs/pythonpath.0.src.2013-07-31_14_30_56/src.tgz#pythonpath.0.src" ^
        #        -files "hdfs:/user/carlk/runs/2013-08-02_13_51_42/distMapper.bat,hdfs:/user/carlk/runs/2013-08-02_13_51_42/distReducer.bat,hdfs:/user/carlk/runs/2013-08-02_13_51_42/distributable.p" ^
        #        -input "runs/2013-08-02_13_51_42/input" ^
        #        -output "runs/2013-08-02_13_51_42/output" ^
        #        -mapper "distMapper.bat" ^
        #       -reducer "distReducer.bat"
        # hadoop fs -cat runs/2013-08-02_13_51_42/output/part-00000  | more
        s00 = r"%HADOOP_HOME%\bin\hadoop fs -rmr -skipTrash {0}".format(
            run_dir_rel.replace("\\", "/")
        )
        s0 = r"%HADOOP_HOME%\bin\hadoop fs -copyFromLocal {0} {1}".format(
            run_dir_rel, run_dir_rel.replace("\\", "/")
        )

        # -D mapreduce.reduce.shuffle.connect.timeout=3600000 ^
        # -D io.sort.mb=1400 ^
        # -D job.end.retry.interval=3600000 ^
        # -D mapred.tasktracker.expiry.interval=3600000 ^

        logging.info("running {0}".format(str(distributable)))

        s = r"""%HADOOP_HOME%\bin\hadoop jar %HADOOP_HOME%\lib\hadoop-streaming.jar ^
        -archives "{0}" ^
        -files "{1}" ^
        -D mapred.job.name="{8}" ^
        -D mapred.map.tasks={4} ^
        -D mapred.reduce.tasks=1 ^
        -D mapred.job.map.memory.mb={5} ^
        -D mapred.job.reduce.memory.mb={6} ^
        -D mapred.task.timeout={7} ^
        -D mapred.job.queue.name="{9}" ^
        -input {2} ^
        -output {3} ^
        -inputformat org.apache.hadoop.mapred.lib.NLineInputFormat ^
        -mapper "distMapper.bat" ^
        -reducer "distReducer.bat"
            """.format(
            ",".join(archivesStringList),  # 0
            filesString,  # 1
            taskIndexDir.replace("\\", "/"),  # 2
            (run_dir_rel + os.path.sep + "output").replace("\\", "/"),  # 3
            self.taskcount,  # 4
            self.mapmemory,  # 5
            self.reducememory,  # 6
            0,  # 7
            str(distributable),  # 8
            self.queue,  # 9
        )
        runHadoopFileName = run_dir_rel + os.path.sep + "runHadoop.bat"
        logging.info("Hadoop runner is creating '{0}'".format(runHadoopFileName))
        with open(runHadoopFileName, "w") as runHadoopFile:
            runHadoopFile.write("call {0}\n".format(s00))
            runHadoopFile.write("call {0}\n".format(s0))
            runHadoopFile.write("call {0}\n".format(s))

        sOneLine = "".join(s.split("^\n"))

        logging.info("Hadoop runner running the copyFromLocal")
        with TemporaryFile() as output:
            stdout0 = subprocess.check_output(s0, stderr=output, shell=True)
            output.seek(0)
            stderr0 = output.read()
        logging.info(
            "Result from 'Hadoop runner running the copyFromLocal' is stdout='{0}', stderr='{1}'".format(
                stdout0, stderr0
            )
        )
        if stderr0 != "":
            raise Exception("Stderr from command: '{0}'".format(stderr0))
        logging.info("Hadoop runner running the streamingjar")

        with TemporaryFile() as output:
            stdout = subprocess.check_output(sOneLine, stderr=output, shell=True)
            output.seek(0)
            stderr = output.read()
        logging.info(
            "Result from 'Hadoop runner running the streamingjar' is stdout='{0}', stderr='{1}'".format(
                stdout, stderr
            )
        )
        logging.info("Done: Hadoop runner is submitting to cluster")
        # if stderr != "" : raise Exception("Stderr from command: '{0}'".format(stderr))

    def FindOrCreatePythonSettings(self, remotewd):
        localpythonpathsetting = r"\\msr-arrays\scratch\msr-pool\eScience3\.continuum"  # os.path.join(os.environ.get("userprofile"),".continuum")
        lastFolderName = os.path.split(os.path.normpath(localpythonpathsetting))[1]
        # pstutil.create_directory_if_necessary(localpythonpathsetting,isfile=False)

        # Must set assume_changed=True for otherwise hidden .continuum file to be used.
        tgzName = HadoopCopier.CheckUpdateTgz(
            localpythonpathsetting,
            subsubItemList1=None,
            skipcheck=False,
            filter_hidden=False,
        )
        hdfstgz = "hdfs:{3}/{2}.{1}/{2}.tgz".format(
            None,
            str(datetime.datetime.fromtimestamp(os.path.getmtime(tgzName)))[:19]
            .replace(" ", "_")
            .replace(":", "_"),
            lastFolderName,
            remotewd,
        )
        Hadoop.hdfsCopyFromLocalIfNotThere(tgzName, hdfstgz)
        return hdfstgz + "#" + lastFolderName

    def create_distributablep(self, distributable, run_dir_abs, run_dir_rel):
        logging.info("Hadoop runner is pickling distributable")
        distributablep_filename_rel = os.path.join(run_dir_rel, "distributable.p")
        # distributablep_filename_abs = os.path.join(run_dir_abs, "distributable.p")
        pstutil.create_directory_if_necessary(distributablep_filename_rel)
        with open(distributablep_filename_rel, mode="wb") as f:
            pickle.dump(distributable, f)
        logging.info("Done: Hadoop runner is pickling distributable")
        return distributablep_filename_rel

    def create_bat_file(
        self,
        distributable,
        remotepythonpath,
        remotewd,
        run_dir_abs,
        run_dir_rel,
        result_remote,
        result_hdfs,
    ):
        logging.info("Hadoop runner is creating bat file")

        outFileList = Hadoop.RecursivelyGetAllOutputs(distributable)

        distributablep_filename_rel = self.create_distributablep(
            distributable, run_dir_abs, run_dir_rel
        )

        distributable_py_file = os.path.join(
            os.path.dirname(__file__), "..", "distributable.py"
        )
        if not os.path.exists(distributable_py_file):
            raise Exception(
                "Expect file at " + distributable_py_file + ", but it doesn't exist."
            )
        localfilepath, file = os.path.split(distributable_py_file)
        remoteexepath = os.path.join(
            remotepythonpath.split(";")[0], "fastlmm", "util"
        )  # !!shouldn't need to assume where the file is in source

        batfilename_abs_list = []
        for part in ["Mapper", "Reducer"]:
            command_string = (
                remoteexepath
                + os.path.sep
                + file
                + r""" distributable.p "Local{0}({1},""{2}"",mkl_num_threads={3},logging_handler=logging.StreamHandler())" """.format(
                    part,
                    self.taskcount,
                    result_remote.replace(
                        "\\", "/"
                    ),  # change to DOS separator to Unix separator because python will work with either and this avoid problems with parsing the batch file
                    self.mkl_num_threads,
                )
            )
            batfilename_rel = os.path.join(run_dir_rel, "dist{0}.bat".format(part))
            batfilename_abs = "hdfs:" + os.path.join(
                run_dir_abs, "dist{0}.bat".format(part)
            ).replace("\\", "/")
            batfilename_abs_list.append(batfilename_abs)
            pstutil.create_directory_if_necessary(batfilename_rel, isfile=True)
            with open(batfilename_rel, "w") as batfile:
                batfile.write(
                    "@set path={0};{0}\Scripts;%path%\n".format(r"c:\GCD\esciencepy")
                )
                batfile.write("@set PYTHONPATH={0}\n".format(remotepythonpath))
                batfile.write("@set home=%cd%\n")
                # batfile.write("@mklink /d .continuum continuum\n")
                # batfile.write("@dir /s\n")
                # batfile.write("@set R_HOME={0}\n".format(os.path.join(remotepythoninstall,"R-2.15.2")))
                # batfile.write("@set R_USER={0}\n".format("."))
                batfile.write(
                    "@mkdir {0}\n@mkdir {0}\\tex.cache\n@set MPLCONFIGDIR={0}\n".format(
                        ".matplotlib"
                    )
                )
                batfile.write("@mkdir {0}\nset IPYTHONDIR={0}\n".format(".ipython"))
                # batfile.write("xcopy /d /e /s /c /h /i continuum .continuum\n")
                batfile.write("@call python {0}\n".format(command_string))
                if part == "Reducer":
                    batfile.write(
                        "@call %HADOOP_HOME%\\bin\\hadoop fs -rm {0} -skipTrash\n".format(
                            result_hdfs
                        )
                    )
                    batfile.write(
                        "@call %HADOOP_HOME%\\bin\\hadoop fs -copyFromLocal {0} {1}\n".format(
                            result_remote, result_hdfs
                        )
                    )
                    for outfile in outFileList:
                        hdfsOutFile = remotewd + "/" + outfile
                        batfile.write(
                            "@call %HADOOP_HOME%\\bin\\hadoop fs -rm {0}\n".format(
                                hdfsOutFile
                            )
                        )
                        batfile.write(
                            "@call %HADOOP_HOME%\\bin\\hadoop fs -copyFromLocal {0} {1}\n".format(
                                outfile, hdfsOutFile
                            )
                        )
        picklefilename_abs = "hdfs:" + os.path.join(
            run_dir_abs, "distributable.p"
        ).replace("\\", "/")
        batfilename_abs_list.append(picklefilename_abs)
        logging.info("Done: Hadoop runner is creating bat file")
        return batfilename_abs_list

    @staticmethod
    def RecursivelyGetAllOutputs(item):
        outputList = []
        ListCopier([], outputList).output(item)
        return outputList

    def FindOrCreateRemotePythonPath(self, localpythonpath, remotewd):
        # input: something like: 'D:\\Source\\carlkextranet05312013\\ERG01\\src\\' and maybe a second item, etc
        # sideeffect: create D:\\Source\\carlkextranet05312013\\ERG01\\src.tgz and 2nd item
        # sideeffect: copy if newer to to hdfs /user/carlk/runs/pythonpath.src.0.<moddate>/src.tgz and ....1.....2nditem
        # return 1 list of "hdfs /user/carlk/runs/pythonpath.0.<moddate>/src.tgz#pythonpath.0.src"
        #       2 remotepythonpath, e.g. "pythonpath.0.src;pythonpath.1.2nditem"

        remotepythonpath_list = []
        tgzList = []
        for i, localpythonpathdir in enumerate(localpythonpath.split(";")):
            tgzName = HadoopCopier.CheckUpdateTgz(
                localpythonpathdir, skipcheck=self.skipsourcecheck
            )
            lastFolderName = os.path.split(os.path.normpath(localpythonpathdir))[1]
            hdfstgz = "hdfs:{3}/pythonpath.{0}.{2}.{1}/{2}.tgz".format(
                i,
                str(datetime.datetime.fromtimestamp(os.path.getmtime(tgzName)))[:19]
                .replace(" ", "_")
                .replace(":", "_"),
                lastFolderName,
                remotewd,
            )
            Hadoop.hdfsCopyFromLocalIfNotThere(tgzName, hdfstgz)
            remotepythonpathdir = "pythonpath.{0}.{1}".format(i, lastFolderName)
            remotepythonpath_list.append(remotepythonpathdir)
            tgzList.append(hdfstgz + "#" + remotepythonpathdir)

        remotepythonpath = ";".join(remotepythonpath_list)

        # these are parallel
        return remotepythonpath, tgzList

    @staticmethod
    def hdfsCopyFromLocalIfNotThere(tgzName, hdfstgz):
        # if it is there won't copy. If it isn't there will copy.
        subprocess.check_output(
            r"type {0} | %HADOOP_HOME%\bin\Hadoop fs -put - {1}".format(
                tgzName, hdfstgz[5:]
            ),
            stderr=subprocess.STDOUT,
            shell=True,
        )

    def create_run_dir(self):
        username = os.environ["USERNAME"]
        localwd = os.getcwd()
        if localwd.startswith("\\\\"):
            remotewd = (
                self.fileshare
                + os.path.sep
                + username
                + os.path.sep
                + "\\".join(localwd.split("\\")[4:])
            )
        else:
            remotewd = (
                self.fileshare + os.path.sep + username + os.path.splitdrive(localwd)[1]
            )  # using '+' because 'os.path.join' isn't work with shares
        remotewd = remotewd.replace("\\", "/")
        if remotewd.endswith("/"):  # remove trailing /
            remotewd = remotewd[:-1]
        run_dir_rel = os.path.join("runs", pstutil._datestamp(appendrandom=True))
        pstutil.create_directory_if_necessary("runs", isfile=False)
        if not os.path.isfile(".ignoreTgzChange"):
            with open("runs" + os.path.sep + ".ignoreTgzChange", "w") as ignoreFile:
                ignoreFile.write("\n")

        run_dir_abs = "/user/{0}/{1}".format(username, run_dir_rel)
        # !! hadoop_create_directory_if_necessary(run_dir_abs,isfile=False)
        return remotewd, run_dir_abs, run_dir_rel

    # !! move these hadoop commands to a library
    @staticmethod
    def hadoop_create_directory_if_necessary(name, isfile=True):
        import os

        if isfile:
            directory_name = os.path.dirname(name)
        else:
            directory_name = name
        hadoop_makedirs(directory_name)

    # !! what if already is there?
    @staticmethod
    def hadoop_makedirs(directory_name):
        hadoop_command("fs -mkdir {0}".format(directory_name))

    @staticmethod
    def hadoop_command(command_string):
        rc = os.system(
            os.environ.get("HADOOP_HOME")
            + os.path.sep
            + "bin"
            + os.path.sep
            + "hadoop "
            + command_string
        )
        if rc != 0:
            raise Exception(
                "Hadoop command '{0}' fails with rc={1}.".format(command_string, rc)
            )


class ListCopier(object):  # Implements ICopier
    def __init__(self, inputList, outputList):  # The list will be modified
        if len(inputList) != 0:
            raise Exception("Expect inputList to start empty")
        if len(outputList) != 0:
            raise Exception("Expect outputList to start empty")
        self.inputList = inputList
        self.outputList = outputList

    def input(self, item):
        if isinstance(item, str):
            self.inputList.append(item)
        elif hasattr(item, "copyinputs"):
            item.copyinputs(self)
            # else do nothing
            pass  # ignore

    def output(self, item):
        if isinstance(item, str):
            self.outputList.append(item)
        elif hasattr(item, "copyoutputs"):
            item.copyoutputs(self)
            # else do nothing
            pass  # ignore


class HadoopCopier(object):  # Implements ICopier
    def __init__(
        self, remotewd, fileInWorkingDirectoryList, tgzList, skipdatacheck
    ):  # The two lists will be modified
        if len(fileInWorkingDirectoryList) != 0:
            raise Exception("Expect fileInWorkingDirectoryList to start empty")
        if len(tgzList) != 0:
            raise Exception("Expect tgzList to start empty")
        self.remotewd = remotewd
        self.fileInWorkingDirectoryList = fileInWorkingDirectoryList
        self.tgzList = tgzList
        self.skipdatacheck = skipdatacheck

    def input(self, item):
        inputList = self.RecursivelyGetAllInputs(item)

        # group by subfolder, treating files in the working directory as special and giving an error with higher levels
        (
            fileInWorkingDirectoryList,
            subDirectoryToSubSubItemList,
        ) = self.GroupByTopLevelSubFolder(inputList)

        # create or update a tgz for each directory
        for directory, subsubItemList in subDirectoryToSubSubItemList.iteritems():
            tgzName = HadoopCopier.CheckUpdateTgz(
                directory, subsubItemList1=subsubItemList, skipcheck=self.skipdatacheck
            )
            hdfsName = "{0}/{1}.{2}.tgz".format(
                self.remotewd,
                directory,
                str(datetime.datetime.fromtimestamp(os.path.getmtime(tgzName)))[:19]
                .replace(" ", "_")
                .replace(":", "_"),
            ).replace("\\", "/")
            self.tgzList.append((tgzName, hdfsName))
            hdfsNameWild = "{0}/{1}.{2}.tgz".format(
                self.remotewd, directory, "*"
            ).replace("\\", "/")
            lookout = subprocess.check_output(
                r"%HADOOP_HOME%\bin\Hadoop fs -ls {0}".format(hdfsName),
                stderr=subprocess.STDOUT,
                shell=True,
            )
            if "No such file or directory" in lookout:
                subprocess.check_output(
                    r"%HADOOP_HOME%\bin\Hadoop fs -rm {0}".format(hdfsNameWild),
                    stderr=subprocess.STDOUT,
                    shell=True,
                )
                subprocess.check_output(
                    r"%HADOOP_HOME%\bin\Hadoop fs -copyFromLocal {0} {1}".format(
                        tgzName, hdfsName
                    ),
                    stderr=subprocess.STDOUT,
                    shell=True,
                )

        for file in fileInWorkingDirectoryList:
            if os.path.getsize(file) > 10 * 1024 * 1024:
                logging.warn(
                    "File '{0}' is large ({1}GB) and would be more efficiently  distributed on Hadoop if placed in subfolder of the working directory.".format(
                        file, round(os.path.getsize(file) / (1024 * 1024))
                    )
                )
            self.fileInWorkingDirectoryList.append(file)
            # filesInWorkingDirectory copy to hdfs

    @staticmethod
    def CheckUpdateTgz(
        directory0, subsubItemList1=None, skipcheck=False, filter_hidden=True
    ):
        if subsubItemList1 == None:
            subsubItemList1 = [[]]
        directory = os.path.normpath(directory0)
        tgzName = directory + ".tgz"

        if (
            not skipcheck
            and os.path.exists(tgzName)
            and not [] in subsubItemList1
            and not filter_hidden
        ):
            if not os.path.isfile(tgzName):
                raise Exception("Expect '{0}' to be a file.".format(tgzName))
            logging.info("Making list of any files already in {0}".format(tgzName))
            tarfile = tarfileLibrary.open(tgzName, "r")
            for tarinfo in tarfile.getmembers():
                if tarinfo.isfile():
                    filenamein = tarinfo.filename.replace(
                        "/", os.path.sep
                    )  # "/" is the tar separator
                    filename = os.path.join(directory, filenamein)
                    if os.path.exists(filename) and os.path.isfile(filename):
                        subsubItemList1.append(
                            filenamein.split("/")
                        )  # "/" is the tar separator
            tarfile.close()

        subsubItemListOut = HadoopCopier.RemoveRedundant(subsubItemList1)

        if not HadoopCopier.IsTarFileUpToDate(
            directory, subsubItemListOut, tgzName, filter_hidden
        ):
            HadoopCopier.CreateNewTarFile(
                directory, subsubItemListOut, tgzName, filter_hidden
            )
        return tgzName

    @staticmethod
    def CreateNewTarFile(directory, subsubItemList, tgzName, filter_hidden=True):
        # logging.info("{0}, {1}, {2}".format(directory, subsubItemList, tgzName))
        directory1 = os.path.normpath(directory)

        tgzNameTemp = (
            tgzName + ".tar"
        )  # create a temp file so that both files will exist for a while
        logging.info("Creating '{0}'".format(tgzNameTemp))
        tarfile = tarfileLibrary.open(tgzNameTemp, "w:")
        for subsubItem in subsubItemList:
            tarName = "/".join(subsubItem)  # tar files use "/" not os.path.sep
            tarfile.add(
                os.path.normpath(directory + "/" + tarName),
                tarName,
                recursive=True,
                filter=lambda x, directory1=directory1, filter_hidden=filter_hidden: HadoopCopier.tarfile_filter_hidden(
                    x, directory1, filter_hidden
                ),
            )
        tarfile.close()
        logging.info("Compressing '{0}'".format(tgzNameTemp))
        subprocess.call(
            r"c:\cygwin64\bin\gzip.exe --force --fast {0}".format(tgzNameTemp),
            shell=True,
        )
        logging.info("Finished Compressing '{0}'".format(tgzNameTemp))
        if os.path.exists(tgzName):
            os.remove(tgzName)
        os.rename(tgzNameTemp + ".gz", tgzName)

    @staticmethod
    def RemoveRedundant(subsubItemList1):
        subsubItemListOut = []
        for index1, subsubItem1 in enumerate(subsubItemList1):
            if HadoopCopier.shouldinclude(index1, subsubItem1, subsubItemList1):
                subsubItemListOut.append(subsubItem1)
        return subsubItemListOut

    @staticmethod
    def shouldinclude(index1, subsubItem1, subsubItemList1):
        for index2, subsubItem2 in enumerate(subsubItemList1):
            if HadoopCopier.knocksout(index1, subsubItem1, index2, subsubItem2):
                return False
        return True

    @staticmethod
    def knocksout(index1, subsubItem1, index2, subsubItem2):
        if index1 == index2:
            return False
        if len(subsubItem1) == len(subsubItem2) and index1 > index2:
            return False
        if len(subsubItem1) >= len(subsubItem2):
            for i in xrange(len(subsubItem2)):
                if subsubItem1[i] != subsubItem2[i]:
                    return False
            return True
        return False

    def RecursivelyGetAllInputs(self, item):
        inputList = []
        ListCopier(inputList, []).input(item)
        return inputList

    def GroupByTopLevelSubFolder(self, inputList):
        filesInWorkingDirectory = []
        subDirectoryToSubSubItemList = defaultdict(list)
        for item in inputList:
            normpath = os.path.normpath(item)
            try:
                relpath = os.path.relpath(normpath)
            except Exception:
                raise Exception(
                    "for Hadoop input files must be in or below the current directory ('{0}' is not)".format(
                        item
                    )
                )

            if not os.path.exists(relpath):
                raise Exception(
                    "Expect input file (or directory) '{0}' but it doesn't exist".format(
                        relpath
                    )
                )
            parts = relpath.split(os.path.sep)
            if os.path.isfile(relpath):  # If the input item is a file ...
                if len(parts) == 1:  # If the input item is in the working directory
                    filesInWorkingDirectory.append(
                        relpath
                    )  # add it to the filesInWorkingDirectory list
                else:  # A file in a subfolder
                    subDirectoryToSubSubItemList[parts[0]].append(
                        parts[1:]
                    )  # Add it to the list for the subfolder
            else:  # A folder
                if not os.path.isdir(relpath):
                    raise Exception("assert")
                if len(parts) == 1:
                    subDirectoryToSubSubItemList[relpath].append([])
                else:
                    subDirectoryToSubSubItemList[parts[0]].append(parts[1:])
        return filesInWorkingDirectory, subDirectoryToSubSubItemList

    @staticmethod
    def IsTarFileUpToDate(directory, subsubItemList, tgzName, filter_hidden):
        if os.path.exists(tgzName):
            logging.info("Checking if '{0}' is up to date".format(tgzName))
            if not os.path.isfile(tgzName):
                raise Exception("Expect '{0}' to be a file.".format(tgzName))
            tarfile = tarfileLibrary.open(tgzName, "r")
            isUpToDate = HadoopCopier.IsTarFileUpToDateInternal(
                directory, subsubItemList, tarfile, filter_hidden
            )
            tarfile.close()
            logging.info("'{0}' up to date? {1} ".format(tgzName, isUpToDate))
            return isUpToDate
        else:
            return False

    @staticmethod
    def IsTarFileUpToDateInternal(directory, subsubItemList, tgzFile, filter_hidden):
        howToIgnoreString = "To ignore changes in a directory add a file '{0}'".format(
            HadoopCopier._ignoreTgzChangeFileName
        )

        for subsubItem in subsubItemList:
            tarName = "/".join(
                subsubItem
            )  # use "/" instead of os.path.sep because that is what the tar file uses
            winfileOrDirectory = os.path.normpath(directory + os.path.sep + tarName)
            try:
                member = tgzFile.getmember(tarName)
            except Exception as e:
                logging.info(
                    "'{0}' not up to date because of exception {1}. ({2})".format(
                        tarName, e, howToIgnoreString
                    )
                )
                return False
            else:
                if os.path.isfile(winfileOrDirectory):
                    if int(os.path.getmtime(winfileOrDirectory)) != int(member.mtime):
                        logging.info(
                            "'{0}' not up to date because of date change in '{1}'. ({2})".format(
                                tgzFile, winfileOrDirectory, howToIgnoreString
                            )
                        )
                        return False
                else:
                    if not os.path.isdir(winfileOrDirectory):
                        raise Exception(
                            "Expect '{0}' to be a directory. ({1})".format(
                                winfileOrDirectory, howToIgnoreString
                            )
                        )
                    for winfileOrDirectory2 in HadoopCopier.WalkWithoutHidden(
                        winfileOrDirectory, filter_hidden
                    ):
                        tarName2 = os.path.relpath(
                            winfileOrDirectory2, winfileOrDirectory
                        ).replace("\\", "/")
                        try:
                            member = tgzFile.getmember(tarName2)
                        except KeyError:
                            logging.info(
                                "'{0}' not up to date because '{1}' is not found. ({2})".format(
                                    tgzFile, tarName2, howToIgnoreString
                                )
                            )
                            return False
                        else:
                            if os.path.isfile(winfileOrDirectory2):
                                if int(os.path.getmtime(winfileOrDirectory2)) != int(
                                    member.mtime
                                ):
                                    logging.info(
                                        "'{0}' not up to date because of date change in '{1}'. ({2})".format(
                                            tgzFile,
                                            winfileOrDirectory2,
                                            howToIgnoreString,
                                        )
                                    )
                                    return False
        return True

    @staticmethod
    def WalkWithoutHidden(directory, filter_hidden):
        if not HadoopCopier.is_hidden(directory, filter_hidden):
            for sub in os.listdir(directory):
                subfull = os.path.join(directory, sub)
                if os.path.isfile(subfull):
                    if not HadoopCopier.is_hidden(subfull, filter_hidden):
                        yield subfull
                else:
                    for file in HadoopCopier.WalkWithoutHidden(subfull, filter_hidden):
                        yield file

    @staticmethod
    def dont_filter(tarinfo, directory1):
        filenamein = tarinfo.name.replace("/", os.path.sep)  # "/" is the tar separator
        filename = os.path.join(directory1, filenamein)
        logging.info("Adding file to tar '{0}'".format(filename))
        return tarinfo

    @staticmethod
    def tarfile_filter_hidden(tarinfo, directory1, filter_hidden):
        filenamein = tarinfo.name.replace("/", os.path.sep)  # "/" is the tar separator
        filename = os.path.join(directory1, filenamein)
        if HadoopCopier.is_hidden(filename, filter_hidden):
            # logging.info("skipping '{0}'".format(filenamein))
            return None
        else:
            logging.info("Adding file to tar '{0}'".format(filename))
            return tarinfo

    @staticmethod
    def is_hidden(filepath, filter_hidden):
        if not filter_hidden:
            return False
        else:
            name = os.path.basename(os.path.abspath(filepath))
            return (
                name.startswith(".")
                or HadoopCopier.has_hidden_attribute(filepath)
                or HadoopCopier.containsDotIgnoreTgzChange(filepath)
            )

    _ignoreTgzChangeFileName = ".ignoreTgzChange"

    # !! test that this stops it from look down below
    @staticmethod
    def containsDotIgnoreTgzChange(filepath):
        signalPath = os.path.join(filepath, HadoopCopier._ignoreTgzChangeFileName)
        return os.path.exists(signalPath)

    @staticmethod
    def has_hidden_attribute(filepath):
        try:
            attrs = ctypes.windll.kernel32.GetFileAttributesW(unicode(filepath))
            assert attrs != -1
            result = bool(attrs & 2)
        except (AttributeError, AssertionError):
            result = False
        return result

    def output(self, item):
        outFileList = Hadoop.RecursivelyGetAllOutputs(item)
        for outfile in outFileList:
            if os.path.exists(outfile):
                os.remove(outfile)
            hdfsOutFile = self.remotewd + "/" + outfile
            subprocess.check_output(
                "%HADOOP_HOME%\\bin\\hadoop fs -copyToLocal {0} {1}\n".format(
                    hdfsOutFile, outfile
                ),
                stderr=subprocess.STDOUT,
                shell=True,
            )



from __future__ import absolute_import
import os
import shutil
import logging
import tempfile


class FileCache(object):
    """
    A FileCache is class such as :class:`.LocalCache` or :class:`.PeerToPeer` such that
       * for reading, copies a (possibly remote) file to a local disk. (If the local file already exists
         and is up-to-date, retrieval is typically skipped.)
       * for writing, copies a newly-written local file to (possibly remote) storage.
       * provides an assortment of file- and directory-related commands.

    :Example:

        Suppose all machines in a cluster can read & write the storage at 'peertopeer1/common'.
        Also, suppose that 'peertopeer1/192.168.1.105' is stored locally
        on machine 192.168.1.105 but readable other machines in the cluster
        (and so on for all the machines and their IP addresses).

        >>> from pysnptools.util.filecache import PeerToPeer, ip_address
        >>> def id_and_path_function():
        ...     ip = ip_address()
        ...     return ip, 'peertopeer1/{0}'.format(ip)
        >>> file_cache = PeerToPeer(common_directory='peertopeer1/common',id_and_path_function=id_and_path_function)
        >>> file_cache
        PeerToPeer('peertopeer1/common',id_and_path_function=...')

        Remove anything already in remote storage

        >>> file_cache.rmtree()

        Create a random SNP file and store it remotely.

        >>> from pysnptools.snpreader import SnpGen, Dense
        >>> snp_gen = SnpGen(seed=123,iid_count=100,sid_count=500)
        >>> with file_cache.open_write('r123.100x500.dense.txt') as local_filename:
        ...    dense1 = Dense.write(local_filename,snp_gen.read())
        >>> list(file_cache.walk())
        ['r123.100x500.dense.txt']

        Copy back from remote storage (if needed) and then read SNP data from local file.

        >>> with file_cache.open_read('r123.100x500.dense.txt') as local_filename:
        ...     dense2 = Dense(local_filename)
        ...     print(dense2[:3,:3].read().val) #Read 1st 3 individuals and SNPs
        [[ 0. nan nan]
         [ 0.  0. nan]
         [ 0.  0.  0.]]

    Given an appropriate module, such as :class:`.LocalCache` or :class:`.PeerToPeer`, the `FileCache` library provides a unified way
    to work with any remote storage scheme. It differs from virtual file systems, such as CIFS VFS, because:

        * `FileCache` works with all operating systems and requires no operating system changes.
        * `FileCache` can take advantage of the highest performance file retrieval methods (e.g. kernel-space file systems, peer-to-peer transfers, tree copies, etc).
        * `FileCache` can take advantage of the highest performance local read and write storage (e.g. SSDs)
        * `FileCache` can work on top of CIFS VFS and any other remote storage system with an appropriate module.

        The downside of `FileCache` is that:

        * Writing generally requires two statements (e.g. :meth:`open_write`, :meth:`.Dense.write`) instead of just one. Likewise,
          Reading generally requires two statements (e.g. :meth:`open_read`, :meth:`.SnpReader.read`) instead of just one.


    Methods & Properties:

        Every `FileCache`, such as :class:`.LocalCache` and :class:`.PeerToPeer`, has this property:
        :attr:`name`,
        and these methods:
        :meth:`file_exists`, :meth:`getmtime`, :meth:`join`, :meth:`load`, :meth:`open_read`, :meth:`open_write`, :meth:`remove`, :meth:`rmtree`,
        :meth:`save`, and :meth:`walk`. See below for details.

    Details of Methods & Properties:

    """

    def __init__(self):
        super(FileCache, self).__init__()

    @staticmethod
    def _fixup(cache_value, default_subfolder="filecache"):
        if isinstance(cache_value, FileCache):
            return cache_value
        if cache_value is None:
            dirpath = tempfile.mkdtemp() + "/" + default_subfolder
            return FileCache._fixup(dirpath)
        if isinstance(cache_value, str):
            from pysnptools.util.filecache import LocalCache

            return LocalCache(cache_value)
        raise Exception(
            "Do not know how to make a FileCache from '{0}'".format(cache_value)
        )

    def join(self, path):
        """
        The :class:`FileCache` created by appending a path to the current :class:`FileCache`.

        :param path: path to join to current :class:`FileCache`
        :type path: string

        :rtype: :class:`FileCache`

        >>> from pysnptools.util.filecache import LocalCache
        >>> file_cache = LocalCache('localcache1')
        >>> file_cache
        LocalCache('localcache1')
        >>> sub = file_cache.join('sub1')
        >>> sub
        LocalCache('localcache1/sub1')

        """
        head = self._normpath(path)
        if head is None:
            return self
        return self._simple_join(head)

    def walk(self, path=None):
        """
        Generates the relative paths of the files in the :class:`FileCache`. It is OK if there are no files.

        :param path: (Default, None, the current :class:`FileCache`). Optional path (subdirectory, not file) to start in.
        :type path: string

        :rtype: a generator of strings

        :Example:

        >>> from pysnptools.util.filecache import LocalCache
        >>> file_cache = LocalCache('localcache1')
        >>> file_cache.rmtree()
        >>> list(file_cache.walk())
        []

        >>> file_cache.save('file1.txt','Hello')
        >>> list(file_cache.walk())
        ['file1.txt']

        >>> list(file_cache.walk('no_such_sub_path'))
        []
        """
        for item in self.join(path)._simple_walk():
            yield item if path is None else path + "/" + item

    def rmtree(self, path=None, updater=None):
        """
        Delete all files in this :class:`FileCache`. It is OK if there are no files.

        :param path: (Default, None, the current :class:`FileCache`). Optional a path (subdirectory, not file) to start in.
        :type path: string

        :param updater: (Default, None). Optional function to which status messages may be written. For example,
            the function created by :func:`.log_in_place`.
        :type updater: function or lambda
        """
        self.join(path)._simple_rmtree(updater=updater)

    def file_exists(self, file_name):
        """
        Tells if there a file with this name. (A directory with the name doesn't count.)

        :param file_name: The name of the file to look for
        :type file_name: string

        :rtype: bool

        :Example:

        >>> from pysnptools.util.filecache import LocalCache
        >>> file_cache = LocalCache('localcache1')
        >>> file_cache.rmtree()
        >>> file_cache.file_exists('localcache1/file1.txt')
        False

        >>> file_cache.save('localcache1/file1.txt','Hello')
        >>> file_cache.file_exists('localcache1/file1.txt')
        True
        """
        directory, simple_file = self._split(file_name)
        return directory._simple_file_exists(simple_file)

    def open_read(self, file_name, updater=None):
        """
        Used with a 'with' statement to produce a local copy of a (possibly remote) file.

        :param file_name: The name of the (possibly remote) file to read
        :type file_name: string

        :param updater: (Default, None). Optional function to which status messages may be written. For example,
            the function created by :func:`.log_in_place`.
        :type updater: function or lambda

        :rtype: a local filename to read.

        :Example:

        >>> from pysnptools.util.filecache import LocalCache
        >>> file_cache = LocalCache('localcache1')
        >>> file_cache.rmtree()
        >>> file_cache.save('file1.txt','Hello')
        >>> with file_cache.open_read('file1.txt') as local_filename:
        ...     with open(local_filename,'r') as fp:
        ...         line = fp.readline()
        >>> line
        'Hello'

        """
        directory, simple_file_name = self._split(file_name)
        return directory._simple_open_read(simple_file_name, updater=updater)

    def open_write(self, file_name, size=0, updater=None):
        """
        Used with a 'with' statement to produce a local file name that will be copied to remote storage
        when the 'with' statement is exited.

        :param file_name: The name of the (possibly remote) file to which to write.
        :type file_name: string

        :param size: (default 0) if given, an error will be thrown immediately if local storage doesn't
            have room for that many bytes.
        :type size: number

        :param updater: (Default, None). Optional function to which status messages may be written. For example,
            the function created by :func:`.log_in_place`.
        :type updater: function or lambda

        :rtype: a local filename to read.

        :Example:

        >>> from pysnptools.util.filecache import LocalCache
        >>> file_cache = LocalCache('localcache1')
        >>> file_cache.rmtree()
        >>> with file_cache.open_write('file1.txt') as local_filename:
        ...     with open(local_filename,'w') as fp:
        ...         _= fp.write('Hello')
        >>> file_cache.load('file1.txt')
        'Hello'

        """
        directory, simple_file_name = self._split(file_name)
        return directory._simple_open_write(
            simple_file_name, size=size, updater=updater
        )

    def remove(self, file_name, updater=None):
        """
        Remove a file from storage. It is an error to remove a directory this way.

        :param file_name: The name of the (possibly remote) file to remove.
        :type file_name: string

        :param updater: (Default, None). Optional function to which status messages may be written. For example,
            the function created by :func:`.log_in_place`.
        :type updater: function or lambda

        """
        directory, simple_file = self._split(file_name)
        return directory._simple_remove(simple_file, updater=updater)

    def save(self, file_name, contents, size=0, updater=None):
        """
        Write a string to a file in storage.

        :param file_name: The name of the (possibly remote) file to which to write.
        :type file_name: string

        :param contents: What to write to the file.
        :type contents: string

        :param size: (default 0) if given, an error will be thrown immediately if local storage doesn't
            have room for that many bytes.
        :type size: number

        :param updater: (Default, None). Optional function to which status messages may be written. For example,
            the function created by :func:`.log_in_place`.
        :type updater: function or lambda


        >>> from pysnptools.util.filecache import LocalCache
        >>> file_cache = LocalCache('localcache1')
        >>> file_cache.rmtree()
        >>> file_cache.save('file1.txt','Hello')
        >>> file_cache.load('file1.txt')
        'Hello'

        """
        with self.open_write(file_name, size=size, updater=updater) as local_file_name:
            with open(local_file_name, "w") as fp:
                fp.write(contents)

    def load(self, file_name, updater=None):
        """
        Returns the contents of a file in storage as a string.

        :param file_name: The name of the (possibly remote) file to read
        :type file_name: string

        :param updater: (Default, None). Optional function to which status messages may be written. For example,
            the function created by :func:`.log_in_place`.
        :type updater: function or lambda

        :rtype: string - what was written in the file.

        >>> from pysnptools.util.filecache import LocalCache
        >>> file_cache = LocalCache('localcache1')
        >>> file_cache.rmtree()
        >>> file_cache.save('file1.txt','Hello')
        >>> file_cache.load('file1.txt')
        'Hello'

        """
        with self.open_read(file_name, updater=updater) as local_file_name:
            with open(local_file_name, "r") as fp:
                whole = fp.read()
        return whole

    def getmtime(self, file_name):
        """
        Return the time that the file was last modified.

        :param file_name: The name of the (possibly remote) file of interest
        :type file_name: string

        :rtype: number, the number of seconds since the epoch

        """
        directory, simple_file = self._split(file_name)
        return directory._simple_getmtime(simple_file)

    @property
    def name(self):
        """
        A path-like name for this `FileCache`.

        :rtype: string

        >>> from pysnptools.util.filecache import LocalCache
        >>> file_cache = LocalCache('localcache1')
        >>> file_cache.name
        'localcache1'

        """
        return "FileCache"

    def _split(self, file_name):
        head, tail = os.path.split(file_name)
        if not tail:
            assert tail, "Expect a file name"
        if head == "":
            return self, tail
        else:
            return self.join(head), tail

    def _normpath(self, path):
        if path is None:
            return None
        head = os.path.normpath(path).replace("\\", "/")
        if head == ".":
            return None
        if os.name == "nt":
            assert len(head) < 2 or not head.startswith(r"\\"), "Should not be UNC"
            assert os.path.splitdrive(head)[0] == "", "Should have a drive"
        assert head != ".." and not head.startswith("../"), "Should not leave parent"
        assert not head.startswith("/"), "Should not start with '/'"
        return head

    def _at_least_one(self, sequence):
        for item in sequence:
            return True
        return False

    @staticmethod
    def _create_directory(local):
        import pysnptools.util as pstutil  # put here to avoid recursive nesting

        if os.path.exists(local):
            if os.path.isfile(local):
                os.remove(local)
            else:
                shutil.rmtree(local)
        directory_name = os.path.dirname(local)
        if os.path.exists(directory_name) and os.path.isfile(directory_name):
            os.remove(directory_name)
        pstutil.create_directory_if_necessary(local, isfile=True)


if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO)

    if False:
        from pysnptools.util.filecache import PeerToPeer, ip_address

        def id_and_path_function():
            ip = ip_address()
            return ip, "peertopeer1/{0}".format(ip)

        file_cache = PeerToPeer(
            common_directory="peertopeer1/common",
            id_and_path_function=id_and_path_function,
        )
        file_cache
        # PeerToPeer('peertopeer1/common',id_and_path_function=...')
        file_cache.rmtree()

        from pysnptools.snpreader import SnpGen, Dense

        snp_gen = SnpGen(seed=123, iid_count=1000, sid_count=5000)
        with file_cache.open_write("r123.1000x5000.dense.txt") as local_filename:
            Dense.write(local_filename, snp_gen.read())
        list(file_cache.walk())
        # ['r123.1000x5000.dense.txt']

    import doctest

    doctest.testmod(optionflags=doctest.ELLIPSIS)
    # There is also a unit test case in 'pysnptools\test.py' that calls this doc test



from pysnptools.distreader import DistReader
from pysnptools.pstreader._subset import _PstSubset

class _DistSubset(_PstSubset,DistReader):
    def __init__(self, *args, **kwargs):
        super(_DistSubset, self).__init__(*args, **kwargs)



import logging
import numpy as np
from pysnptools.distreader import DistReader
from pysnptools.pstreader import PstHdf5
import warnings

class DistHdf5(PstHdf5,DistReader):
    '''
    A :class:`.DistReader` for reading \*.dist.hdf5 files from disk.

    See :class:`.DistReader` for general examples of using DistReaders.

    The general HDF5 format is described `here <http://www.hdfgroup.org/HDF5/>`__. The DistHdf5 format stores
    val, iid, sid, and pos information in Hdf5 format.
   
    **Constructor:**
        :Parameters: * **filename** (*string*) -- The DistHdf5 file to read.

        :Example:

        >>> from pysnptools.distreader import DistHdf5
        >>> from pysnptools.util import example_file # Download and return local file name
        >>> hdf5_file = example_file("pysnptools/examples/toydata.snpmajor.dist.hdf5")
        >>> data_on_disk = DistHdf5(hdf5_file)
        >>> print((data_on_disk.iid_count, data_on_disk.sid_count))
        (25, 10000)

    **Methods beyond** :class:`.DistReader`

    '''
    def __init__(self, *args, **kwargs):
        super(DistHdf5, self).__init__(*args, **kwargs)


    @property
    def row(self):
        self._run_once()
        if self._row.dtype.type is not np.str_:
            self._row = np.array(self._row,dtype='str')
        return self._row

    @property
    def col(self):
        self._run_once()
        if self._col.dtype.type is not np.str_:
            self._col = np.array(self._col,dtype='str')
        return self._col

    @staticmethod
    def write(filename, distdata, hdf5_dtype=None, sid_major=True):
        """Writes a :class:`DistData` to DistHdf5 format and return a the :class:`.DistHdf5`.

        :param filename: the name of the file to create
        :type filename: string
        :param distdata: The in-memory data that should be written to disk.
        :type distdata: :class:`DistData`
        :param hdf5_dtype: None (use the .val's dtype) or a Hdf5 dtype, e.g. 'f8','f4',etc.
        :type hdf5_dtype: string
        :param sid_major: Tells if vals should be stored on disk in sid_major (default) or iid_major format.
        :type col_major: bool
        :rtype: :class:`.DistHdf5`

        >>> from pysnptools.distreader import DistHdf5, Bgen
        >>> import pysnptools.util as pstutil
        >>> from pysnptools.util import example_file # Download and return local file name
        >>> bgen_file = example_file("pysnptools/examples/2500x100.bgen")
        >>> distdata = Bgen(bgen_file)[:,:10].read()     # Read first 10 snps from BGEN format
        >>> pstutil.create_directory_if_necessary("tempdir/toydata10.dist.hdf5")
        >>> DistHdf5.write("tempdir/toydata10.dist.hdf5",distdata)        # Write data in DistHdf5 format
        DistHdf5('tempdir/toydata10.dist.hdf5')
        """
        PstHdf5.write(filename,distdata,hdf5_dtype=hdf5_dtype,col_major=sid_major)
        return DistHdf5(filename)

if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO)

    import doctest
    doctest.testmod()



import math
import os
import logging

def is_prime(n):
    '''
    Tells (somewhat inefficiently) of a number is prime.

    >>> from pysnptools.util.mapreduce1.examples import is_prime
    >>> is_prime(5)
    True
    >>> is_prime(9)
    False
    '''
    assert n == int(n) and n>1, "Expect integers greater than 1"
    for j in range(2,int(math.sqrt(n))+1):
        if n % j == 0:
            return False
    return True

def prime_search0(start,stop):
    '''
    Iterative algorithm for finding prime numbers in a range, but just testing them one-by-one.

    >>> from pysnptools.util.mapreduce1.examples import prime_search0
    >>> prime_search0(2,10)
    [2, 3, 5, 7]
    '''
    assert start < stop, "start must be less than stop"
    prime_list = []
    for i in range(start,stop):
        if is_prime(i):
            prime_list.append(i)
    return prime_list

def prime_search1(start,stop,runner):
    '''
    Distributed algorithm for finding prime numbers in a range, but just testing each number.

    >>> from pysnptools.util.mapreduce1.examples import prime_search1
    >>> from pysnptools.util.mapreduce1.runner import LocalMultiProc
    >>> prime_search1(2,10,LocalMultiProc(4))
    [2, 3, 5, 7]
    '''
    from pysnptools.util.mapreduce1 import map_reduce

    def mapper(i):
        if is_prime(i):
            return i
        else:
            return None

    def reducer(sequence):
        result = []
        for i in sequence:
            if i is not None:
                result.append(i)
        return result

    return map_reduce(range(start,stop),
                        mapper=mapper,
                        reducer=reducer, #lambda sequence: [i for i in sequence if i is not None], #Filter out the None's
                        name="prime_search1",
                        runner=runner)

if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO)

    import doctest
    doctest.testmod(optionflags=doctest.ELLIPSIS)



import numpy as np
from itertools import *
import pandas as pd
import logging
from pysnptools.snpreader import SnpReader
from pysnptools.standardizer import Unit
from pysnptools.standardizer import Identity
from pysnptools.pstreader import PstData
import warnings
import time
import pysnptools.util as pstutil

class SnpData(PstData,SnpReader):
    """A :class:`.SnpReader` for holding SNP values (or similar values) in-memory, along with related iid and sid information.
    It is usually created by calling the :meth:`.SnpReader.read` method on another :class:`.SnpReader`, for example, :class:`.Bed`. It can also be constructed directly.

    See :class:`.SnpReader` for details and examples.

    **Constructor:**
        :Parameters: * **iid** (an array of string pair) -- The :attr:`SnpReader.iid` information.
                     * **sid** (an array of strings) -- The :attr:`SnpReader.sid` information.
                     * **val** (a 2-D array of floats) -- The SNP values
                     * **pos** (optional, an array of strings) -- The :attr:`SnpReader.pos` information
                     * **name** (optional, string) -- Information to be display about the origin of this data
                     * **copyinputs_function** (optional, function) -- *Used internally by optional clustering code*
                     * **xp** (optional, a Python module or string) -- The array module that controls **val**,
                               for example, 'numpy' or 'cupy'. Defaults to numpy. Can also be set with the
                               'ARRAY_MODULE' environment variable.

        :Example:

        >>> from pysnptools.snpreader import SnpData
        >>> snpdata = SnpData(iid=[['fam0','iid0'],['fam0','iid1']], sid=['snp334','snp349','snp921'], val=[[0.,2.,0.],[0.,1.,2.]])
        >>> print((snpdata.val[0,1], snpdata.iid_count, snpdata.sid_count))
        (2.0, 2, 3)

    **Equality:**

        Two SnpData objects are equal if their four arrays (:attr:`SnpData.val`, :attr:`SnpReader.iid`, :attr:`SnpReader.sid`, and :attr:`SnpReader.pos`)
        are 'array_equal'. (Their 'name' does not need to be the same).
        If either :attr:`SnpData.val` contains NaN, the objects will not be equal. However, :meth:`SnpData.allclose` can be used to treat NaN as
        regular values.

        :Example:

        >>> import numpy as np
        >>> from pysnptools.snpreader import SnpData
        >>> snpdata1 = SnpData(iid=[['fam0','iid0'],['fam0','iid1']], sid=['snp334','snp349','snp921'], val=[[0.,2.,0.],[0.,1.,2.]], pos=[[0,0,0],[0,0,0],[0,0,0]])
        >>> snpdata2 = SnpData(iid=[['fam0','iid0'],['fam0','iid1']], sid=['snp334','snp349','snp921'], val=[[0.,2.,0.],[0.,1.,2.]], pos=[[0,0,0],[0,0,0],[0,0,0]])
        >>> print(snpdata1 == snpdata2) #True, because all the arrays have the same values.
        True
        >>> print(snpdata1.val is snpdata2.val) #False, because the two arrays have different memory.
        False
        >>> snpdata3 = SnpData(iid=[['a','0'],['b','0']], sid=['snp334','snp349','snp921'], val=[[0.,2.,0.],[0.,1.,2.]], pos=[[0,0,0],[0,0,0],[0,0,0]])
        >>> snpdata4 = SnpData(iid=[['fam0','iid0'],['fam0','iid1']], sid=['snp334','snp349','snp921'], val=[[0.,2.,0.],[0.,1.,2.]], pos=[[0,0,0],[0,0,0],[0,0,0]])
        >>> print(snpdata3 == snpdata4) #False, because the iids are different.
        False
        >>> snpdata5 = SnpData(iid=[['fam0','iid0'],['fam0','iid1']], sid=['snp334','snp349','snp921'], val=[[0.,2.,0.],[0.,1.,np.nan]], pos=[[0,0,0],[0,0,0],[0,0,0]])
        >>> snpdata6 = SnpData(iid=[['fam0','iid0'],['fam0','iid1']], sid=['snp334','snp349','snp921'], val=[[0.,2.,0.],[0.,1.,np.nan]], pos=[[0,0,0],[0,0,0],[0,0,0]])
        >>> print(snpdata5 == snpdata6) #False, because the val's contain NaN
        False
        >>> print(snpdata5.allclose(snpdata6)) #True, if we consider the NaN as regular values, all the arrays have the same values.
        True

    **Methods beyond** :class:`.SnpReader`
    """
    def __init__(self, iid, sid, val, pos=None, name=None, parent_string=None, copyinputs_function=None, xp = None, _require_float32_64=True):
        #We don't have a 'super(SnpData, self).__init__()' here because SnpData takes full responsibility for initializing both its superclasses
        xp = pstutil.array_module(xp)
        self._val = None

        if parent_string is not None:
            warnings.warn("'parent_string' is deprecated. Use 'name'", DeprecationWarning)
        self._row = PstData._fixup_input(iid,empty_creator=lambda ignore:np.empty([0,2],dtype='str'),dtype='str')
        self._col = PstData._fixup_input(sid,empty_creator=lambda ignore:np.empty([0],dtype='str'),dtype='str')
        self._row_property = PstData._fixup_input(None,count=len(self._row),empty_creator=lambda count:np.empty([count,0],dtype='str'),dtype='str')
        self._col_property = PstData._fixup_input(pos,count=len(self._col),empty_creator=lambda count:np.full([count, 3], np.nan))
        self._val = PstData._fixup_input_val(val,row_count=len(self._row),col_count=len(self._col),
                                                empty_creator=lambda row_count,col_count:np.empty([row_count,col_count],dtype=np.float64),
                                                _require_float32_64=_require_float32_64,
                                                xp = xp
                                                )
        self._assert_iid_sid_pos(check_val=True)
        self._name = name or parent_string or ""
        self._std_string_list = []
        self._xp = xp

    @property
    def val(self):
        """The 2D NumPy array of floats that represents the values of the SNPs.  You can get or set this property.

        >>> from pysnptools.snpreader import Bed
        >>> from pysnptools.util import example_file # Download and return local file name
        >>> bed_file = example_file("tests/datasets/all_chr.maf0.001.N300.*","*.bed")
        >>> snpdata = Bed(bed_file,count_A1=False)[:5,:].read() #read data for first 5 iids
        >>> print(snpdata.val[4,100]) #print one of the SNP values
        2.0
        """
        return self._val

    @val.setter
    def val(self, new_value):
        self._val = PstData._fixup_input_val(new_value,row_count=len(self._row),col_count=len(self._col),empty_creator=lambda row_count,col_count:np.empty([row_count,col_count],dtype=np.float64),
                                             xp=self._xp)
        self._assert_iid_sid_pos(check_val=True)


    def allclose(self,value,equal_nan=True):
        '''
        :param value: Other object with which to compare.
        :type value: :class:`SnpData`
        :param equal_nan: (Default: True) Tells if NaN in :attr:`SnpData.val` should be treated as regular values when testing equality.
        :type equal_nan: bool

        >>> import numpy as np
        >>> snpdata5 = SnpData(iid=[['fam0','iid0'],['fam0','iid1']], sid=['snp334','snp349','snp921'], val=[[0.,2.,0.],[0.,1.,np.nan]], pos=[[0,0,0],[0,0,0],[0,0,0]])
        >>> snpdata6 = SnpData(iid=[['fam0','iid0'],['fam0','iid1']], sid=['snp334','snp349','snp921'], val=[[0.,2.,0.],[0.,1.,np.nan]], pos=[[0,0,0],[0,0,0],[0,0,0]])
        >>> print(snpdata5.allclose(snpdata6)) #True, if we consider the NaN as regular values, all the arrays have the same values.
        True
        >>> print(snpdata5.allclose(snpdata6,equal_nan=False)) #False, if we consider the NaN as special values, all the arrays are not equal.
        False

        '''
        return PstData.allclose(self,value,equal_nan=equal_nan)

    def train_standardizer(self, apply_in_place, standardizer=Unit(), force_python_only=False,num_thread=None):
        """
        .. deprecated:: 0.2.23
           Use :meth:`standardize` with return_trained=True instead.
        """
        warnings.warn("train_standardizer is deprecated. standardize(...,return_trained=True,...) instead", DeprecationWarning)
        assert apply_in_place, "code assumes apply_in_place"
        self._std_string_list.append(str(standardizer))
        _, trained_standardizer = standardizer.standardize(self, return_trained=True, force_python_only=force_python_only,num_threads=num_threads)
        return trained_standardizer

    #LATER should there be a single warning if Unit() finds and imputes NaNs?
    def standardize(self, standardizer=Unit(), block_size=None, return_trained=False, force_python_only=False, num_threads=None):
        """Does in-place standardization of the in-memory
        SNP data. By default, it applies 'Unit' standardization, that is: the values for each SNP will have mean zero and standard deviation 1.0.
        NaN values are then filled with zero, the mean (consequently, if there are NaN values, the final standard deviation will not be zero.
        Note that, for efficiency, this method works in-place, actually changing values in the ndarray. Although it works in place, for convenience
        it also returns the SnpData.

        :param standardizer: optional -- Specify standardization to be applied. 
             Any :class:`.Standardizer` may be used. Some choices include :class:`.Unit` (default, makes values for each SNP have mean zero and
             standard deviation 1.0) and :class:`.Beta`.
        :type standardizer: :class:`.Standardizer`

        :param block_size: optional -- Deprecated.
        :type block_size: None

        :param return_trained: If true, returns a second value containing a constant :class:`.Standardizer` trained on this data.
        :type return_trained: bool

        :param force_python_only: optional -- If true, will use pure Python instead of faster C++ libraries.
        :type force_python_only: bool

        :param num_threads: optional -- The number of threads with which to standardize data. Defaults to all available
            processors. Can also be set with these environment variables (listed in priority order):
            'PST_NUM_THREADS', 'NUM_THREADS', 'MKL_NUM_THREADS'.
        :type num_threads: None or int

        :rtype: :class:`.SnpData` (standardizes in place, but for convenience, returns 'self')

        >>> from pysnptools.snpreader import Bed
        >>> from pysnptools.util import example_file # Download and return local file name
        >>> bed_file = example_file("tests/datasets/all_chr.maf0.001.N300.*","*.bed")
        >>> snp_on_disk = Bed(bed_file,count_A1=False) # Specify some data on disk in Bed format
        >>> snpdata1 = snp_on_disk.read() # read all SNP values into memory
        >>> print(snpdata1) # Prints the specification for this SnpData
        SnpData(Bed(...tests/datasets/all_chr.maf0.001.N300.bed',count_A1=False))
        >>> print(snpdata1.val[0,0])
        2.0
        >>> snpdata1.standardize() # standardize changes the values in snpdata1.val and changes the specification.
        SnpData(Bed(...tests/datasets/all_chr.maf0.001.N300.bed',count_A1=False),Unit())
        >>> print('{0:.6f}'.format(snpdata1.val[0,0]))
        0.229416
        >>> snpdata2 = snp_on_disk.read().standardize() # Read and standardize in one expression with only one ndarray allocated.
        >>> print('{0:.6f}'.format(snpdata2.val[0,0]))
        0.229416
        """
        self._std_string_list.append(str(standardizer))
        _, trained = standardizer.standardize(self, return_trained=True, force_python_only=force_python_only, num_threads=num_threads)
        if return_trained:
            return self, trained
        else:
            return self

    def _read_kernel(train, standardizer, block_size=None, order='A', dtype=np.float64, force_python_only=False, view_ok=False, return_trained=False, num_threads=None):
        '''
        The method creates a kernel for the in-memory SNP data. It handles these cases
                * No standardization is needed & everything is in memory  OR uses the FROM-DISK method
        '''
        from pysnptools.pstreader import PstReader
        dtype = np.dtype(dtype)

        #Just do a 'python' dot, if no standardization is needed and everything is the right type
        if isinstance(standardizer,Identity) and train.val.dtype == dtype:
            ts = time.time()
            #is_worth_logging = train.val.shape[0] * train.val.shape[1] * test.val.shape[0] > 1e9
            #if is_worth_logging: logging.info("  _read_kernel about to multiply train{0} x test{1}".format(train.val.shape,test.val.shape))
            if order == 'F': #numpy's 'dot' always returns 'C' order
                K = (train.val.dot(train.val.T)).T
            else:
                K = train.val.dot(train.val.T)
            assert PstReader._array_properties_are_ok(K,order,dtype), "internal error: K is not of the expected order or dtype"
            #if is_worth_logging: logging.info("  _read_kernel took %.2f seconds" % (time.time()-ts))
            if return_trained:
                return K, standardizer
            else:
                return K
        else: #Do things the more general SnpReader way.
            return SnpReader._read_kernel(train, standardizer, block_size=block_size, order=order, dtype=dtype, force_python_only=force_python_only,view_ok=view_ok, return_trained=return_trained, num_threads=num_threads)

    def __repr__(self):
        if self._name == "":
            if len(self._std_string_list) > 0:
                s = "{0}({1})".format(self.__class__.__name__,",".join(self._std_string_list))
            else:
                s = "{0}()".format(self.__class__.__name__)
        else:
            if len(self._std_string_list) > 0:
                s = "{0}({1},{2})".format(self.__class__.__name__,self._name,",".join(self._std_string_list))
            else:
                s = "{0}({1})".format(self.__class__.__name__,self._name)
        return s


if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO)

    import doctest
    doctest.testmod(optionflags=doctest.ELLIPSIS)
    # There is also a unit test case in 'pysnptools\test.py' that calls this doc test



import numpy as np
import logging
from pysnptools.pstreader import PstReader


class PstNpz(PstReader):
    """
    A :class:`.PstReader` for reading `*.pst.npz` files from disk.

    See :class:`.PstReader` for general examples of using PstReaders.

    The general NPZ format is described `here <http://docs.scipy.org/doc/numpy/reference/generated/numpy.savez.html>`__. The PstNpz format stores
    val, row, col, row_property, and col_property information in NPZ format.

    **Constructor:**
        :Parameters: * **filename** (*string*) -- The PstNpz file to read.

        :Example:

        >>> from pysnptools.pstreader import PstNpz
        >>> from pysnptools.util import example_file # Download and return local file name
        >>>
        >>> pstnpz_file = example_file('tests/datasets/little.pst.npz')
        >>> data_on_disk = PstNpz(pstnpz_file)
        >>> print(data_on_disk.row_count)
        300

    **Methods beyond** :class:`.NpzReader`

    """

    def __init__(self, filename):
        """
        filename    : string of the name of the npz file.
        """
        super(
            PstNpz, self
        ).__init__()  # We don't send "file name" up because we know about super doesn't want it.
        self._ran_once = False

        self._filename = filename

    def __repr__(self):
        return "{0}('{1}')".format(self.__class__.__name__, self._filename)

    @property
    def row(self):
        self._run_once()
        return self._row

    @property
    def col(self):
        self._run_once()
        return self._col

    @property
    def row_property(self):
        self._run_once()
        return self._row_property

    @property
    def col_property(self):
        self._run_once()
        return self._col_property

    def _run_once(self):
        if self._ran_once:
            return
        self._ran_once = True

        with np.load(
            self._filename, allow_pickle=True
        ) as data:  # !! similar code in epistasis
            if (
                len(data.keys()) == 2 and "arr_0" in data.keys()
            ):  # for backwards compatibility
                self._row = data["arr_0"]
                self._col = self._row
                self._row_property = np.empty((len(self._row), 0))
                self._col_property = np.empty((len(self._col), 0))
            else:
                self._row = data["row"]
                self._col = data["col"]
                if self._row.dtype == self._col.dtype and np.array_equal(
                    self._row, self._col
                ):  # If it's square, mark it so by making the col and row the same object
                    self._col = self._row
                self._row_property = data["row_property"]
                self._col_property = data["col_property"]

        return self

    def copyinputs(self, copier):
        # doesn't need to self._run_once()
        copier.input(self._filename)

    # Most _read's support only indexlists or None, but this one supports Slices, too.
    _read_accepts_slices = True

    def _read(
        self,
        row_index_or_none,
        col_index_or_none,
        order,
        dtype,
        force_python_only,
        view_ok,
        num_threads,
    ):
        dtype = np.dtype(dtype)
        # 'view_ok' doesn't mean anything here because we are always ready fresh from disk.
        # !! could use mmap so only rows of interest are loaded.
        self._run_once()

        # np.load does the right thing and doesn't load 'val' into memory until accessed here.
        with np.load(
            self._filename, allow_pickle=True
        ) as data:  # !! similar code in epistasis
            if (
                len(data.keys()) == 2 and "arr_1" in data.keys()
            ):  # for backwards compatibility
                val = data["arr_1"]
            else:
                val = data["val"]

        val, _ = self._apply_sparray_or_slice_to_val(
            val,
            row_index_or_none,
            col_index_or_none,
            order,
            dtype,
            force_python_only,
            num_threads,
        )
        return val

    @staticmethod
    def write(filename, pstdata):
        """Writes a :class:`PstData` to PstNpz format and returns the :class:`.PstNpz`.

        :param filename: the name of the file to create
        :type filename: string
        :param pstdata: The in-memory data that should be written to disk.
        :type pstdata: :class:`PstData`
        :rtype: :class:`.PstNpz`

        >>> from pysnptools.pstreader import PstData, PstNpz
        >>> import pysnptools.util as pstutil
        >>> data1 = PstData(row=['a','b','c'],col=['y','z'],val=[[1,2],[3,4],[np.nan,6]],row_property=['A','B','C'])
        >>> pstutil.create_directory_if_necessary("tempdir/tiny.pst.npz")
        >>> PstNpz.write("tempdir/tiny.pst.npz",data1)          # Write data in PstNz format
        PstNpz('tempdir/tiny.pst.npz')
        """
        np.savez(
            filename,
            row=pstdata.row,
            col=pstdata.col,
            row_property=pstdata.row_property,
            col_property=pstdata.col_property,
            val=pstdata.val,
        )
        logging.debug("Done writing " + filename)
        return PstNpz(filename)


if __name__ == "__main__":
    import doctest

    logging.basicConfig(level=logging.INFO)

    doctest.testmod(optionflags=doctest.ELLIPSIS)

    # from pysnptools.snpreader.dat import Dat
    # snpreader = Dat(r'../tests/datasets/all_chr.maf0.001.N300.dat')
    # snp_matrix = snpreader.read()
    # print(len(snp_matrix['sid']))
    # snp_matrix = snpreader[:,:].read()
    # print(len(snp_matrix['sid']))
    # sid_index_list = snpreader.sid_to_index(['23_9','23_2'])
    # snp_matrix = snpreader[:,sid_index_list].read()
    # print(",".join(snp_matrix['sid']))
    # snp_matrix = snpreader[:,0:10].read()
    # print(",".join(snp_matrix['sid']))

    # print(snpreader.iid_count)
    # print(snpreader.sid_count)
    # print(len(snpreader.pos))

    # snpreader2 = snpreader[::-1,4]
    # print(snpreader.iid_count)
    # print(snpreader2.sid_count)
    # print(len(snpreader2.pos))

    # snp_matrix = snpreader2.read()
    # print(len(snp_matrix['iid']))
    # print(len(snp_matrix['sid']))

    # snp_matrix = snpreader2[5,:].read()
    # print(len(snp_matrix['iid']))
    # print(len(snp_matrix['sid']))

    # iid_index_list = snpreader2.iid_to_index(snpreader2.iid[::2])
    # snp_matrix = snpreader2[iid_index_list,::3].read()
    # print(len(snp_matrix['iid']))
    # print(len(snp_matrix['sid']))

    # snp_matrix = snpreader[[4,5],:].read()
    # print(len(snp_matrix['iid']))
    # print(len(snp_matrix['sid']))

    # print(snpreader2)
    # print(snpreader[::-1,4])
    # print(snpreader2[iid_index_list,::3])
    # print(snpreader[:,sid_index_list])
    # print(snpreader2[5,:])
    # print(snpreader[[4,5],:])
